# -*- coding: utf-8 -*-
# Copyright 2016-2019 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

"""Test module for `tkwant.leads`"""

import pytest
from pytest import raises
import cmath
import tinyarray as ta
import numpy as np
import kwant
from numpy.testing import (assert_array_almost_equal, assert_almost_equal)

from .. import leads
from .common import (make_square, make_square_with_leads, make_simple_lead,
                     make_complex_lead, check_boundary_hamiltonian)

import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)


def test_make_time_dependent_hopping():

    def phase(a, b, time):
        return time * a + b

    a_ = 2
    b_ = 3
    c_ = 5
    d_ = -2
    time_ = 10
    ii_ = - 2
    jj_ = - 1
    exp_factor = cmath.exp(1j * phase(a=a_, b=b_, time=time_))

    # hopping is a constant
    old_hopping = -5 + 2 * 1j

    new_hopping = leads._make_time_dependent_hopping(phase, old_hopping, system_to_lead=True)
    signature = kwant._common.get_parameters(new_hopping)
    assert signature == ('_site0_', '_site1_', 'a', 'b', 'time')
    assert new_hopping(None, None, a_, b_, time_) == exp_factor * old_hopping

    new_hopping = leads._make_time_dependent_hopping(phase, old_hopping, system_to_lead=False)
    signature = kwant._common.get_parameters(new_hopping)
    assert signature == ('_site0_', '_site1_', 'a', 'b', 'time')
    assert new_hopping(None, None, a_, b_, time_) == exp_factor.conjugate() * old_hopping

    # hopping is callable
    def old_hopping(ii, jj, c, time, d, a):
        return ii + jj + time * c - d * a

    old_hop_eval = old_hopping(ii=ii_, jj=jj_, c=c_, time=time_, d=d_, a=a_)

    new_hopping = leads._make_time_dependent_hopping(phase, old_hopping, system_to_lead=True)
    signature = kwant._common.get_parameters(new_hopping)
    assert signature == ('ii', 'jj', 'c', 'time', 'd', 'a', 'b')
    assert new_hopping(ii_, jj_, c_, time_, d_, a_, b_) == exp_factor * old_hop_eval

    new_hopping = leads._make_time_dependent_hopping(phase, old_hopping, system_to_lead=False)
    signature = kwant._common.get_parameters(new_hopping)
    assert signature == ('ii', 'jj', 'c', 'time', 'd', 'a', 'b')
    assert new_hopping(ii_, jj_, c_, time_, d_, a_, b_) == exp_factor.conjugate() * old_hop_eval


def test_add_voltage():
    N = 3

    def check_square_lead(syst, hop_expect, lat, args=()):
        # check lead cell was added and interface was updated
        new_cell = set(lat(-1, j) for j in range(N))
        # print([site in syst for site in new_cell])
        assert all(site in syst for site in new_cell)
        assert new_cell == set(syst.leads[0].interface)
        # test value
        hops = [(lat(-1, j), lat(0, j)) for j in range(N)]
        check_hops(syst, hops, hop_expect, args)

    def check_hops(syst, hops, hop_expect, args=()):
        for hop in hops:
            for time in (0,):  # (-1, 0, 1):
                got = syst[hop](*(hop + (time,) + args))
                expected = hop_expect(time, *args)
                assert np.allclose(got, expected)
                rhop = tuple(reversed(hop))
                rgot = syst[rhop](*(rhop + (time,) + args))
                rexpected = expected.conjugate()
                assert np.allclose(rgot, rexpected)

    def raises_ValueError(t):
        raise ValueError('problem in phase function')

    def bad_return_value(t):
        return 'nonsense'

    def bad_return_shape(t):
        return np.eye(2)  # expecting scalar

    def phase(t):
        return t

    def simple_hop(t):
        return -1 * cmath.exp(1j * phase(t))  # hopping should be

    # --- input errors
    lat, syst = make_square(N)
    lead = kwant.Builder(kwant.TranslationalSymmetry((-1, 0)))
    lead[(lat(0, j) for j in range(N))] = 2
    lead[lat.neighbors()] = -1
    # hopping > nearest neighbor lead cell
    lead[kwant.builder.HoppingKind((2, 0), lat)] = -1
    interface = [lat(0, j) for j in range(1, N)]
    syst.leads.append(kwant.builder.BuilderLead(lead, interface))
    # add an extra site sticking out and manually attach lead on right
    # interface contains sites in different domains: error!
    syst[lat(N, 0)] = 4
    lead = kwant.Builder(kwant.TranslationalSymmetry((1, 0)))
    lead[(lat(0, j) for j in range(N))] = 2
    lead[lat.neighbors()] = -1
    interface = [lat(N, 0)] + [lat(N - 1, j) for j in range(1, N)]
    syst.leads.append(kwant.builder.BuilderLead(lead, interface))

    with pytest.raises(ValueError) as exc:
        leads.add_voltage(syst, 0, None)
    assert str(exc.value) == 'Phase function is not callable'
    with pytest.raises(ValueError) as exc:
        leads.add_voltage(syst, 0, phase)
    assert 'hopping connects non-neighboring' in str(exc.value)
    with pytest.raises(ValueError) as exc:
        leads.add_voltage(syst, 1, phase)
    assert str(exc.value) == 'Some interface sites belong to different domains'

    # --- test when Hamiltonian is constants
    lat, syst = make_square_with_leads(N)
    added = leads.add_voltage(syst, 0, phase)
    new_hop = (lat(-1, 0), lat(0, 0))  # one new hopping -- system *to* lead
    check_square_lead(syst, simple_hop, lat)
    # check correct sites are added
    assert set(added) == set(lat(-1, j) for j in range(N))
    #  phase function returns numpy scalar
    lat, syst = make_square_with_leads(N)
    leads.add_voltage(syst, 0, lambda time: np.array(phase(time)))
    check_square_lead(syst, simple_hop, lat)
    # user phase function raises error
    lat, syst = make_square_with_leads(N)
    leads.add_voltage(syst, 0, raises_ValueError)
    with pytest.raises(RuntimeError) as ctx:
        syst[new_hop](*(new_hop + (0.,)))
    msg = ('Error while evaluating user-supplied phase function '
           '"raises_ValueError": problem in phase function')
    assert str(ctx.value) == msg
    # user phase function returns nonsense
    lat, syst = make_square_with_leads(N)
    leads.add_voltage(syst, 0, bad_return_value)
    with pytest.raises(RuntimeError) as ctx:
        syst[new_hop](*(new_hop + (5.,)))  # should raise RuntimeError
    msg = ('Error while evaluating user-supplied phase function '
           '"bad_return_value"')
    assert msg in str(ctx.value)

    # --- test when Hamiltonian is functions
    def phase(t, arg):
        return t

    def simple_hop(t, arg):
        return -1 * cmath.exp(1j * phase(t, arg))

    lat, syst = make_square_with_leads(N, functional=True)
    leads.add_voltage(syst, 0, phase=phase)
    check_square_lead(syst, simple_hop, lat, args=(None,))

    # --- matrix onsite structure
    sigma_z = ta.array([[1, 0], [0, -1]])

    def phase_mat(t):
        return t * ta.array((1, -1))

    def mat_hop(t):
        return np.dot(np.diag(np.exp(1j * phase_mat(t))), -sigma_z)

    def trivial_phase(t):
        return t

    def trivial_hop(t):
        return np.dot(cmath.exp(1j * trivial_phase(t)), -sigma_z)

    def test_mat(phase, should_be):
        lat, syst = make_square_with_leads(N, norbs=2,
                                           mat=lambda n: sigma_z)
        leads.add_voltage(syst, 0, phase=phase)
        check_square_lead(syst, should_be, lat)

    # check when return value is 1D sequence -- the diagonal
    test_mat(phase_mat, mat_hop)
    # check when return value is true scalar
    test_mat(trivial_phase, trivial_hop)
    # check when return value is numpy scalar
    test_mat(lambda t: np.array(trivial_phase(t)), trivial_hop)

    # --- nontrivial leads
    def expected(time):
        return np.dot(np.diag(np.exp(1j * phase_mat(time))), 1j * sigma_z)

    lat, syst = make_square(N, norbs=2)
    lead = kwant.Builder(kwant.TranslationalSymmetry((-2, 0)))
    tags = ((0, 1), (1, 2), (1, 0))
    lead[(lat(*t) for t in tags)] = 4 * sigma_z
    lead[lat.neighbors(2)] = -np.eye(2)
    lead2 = kwant.Builder(kwant.TranslationalSymmetry((2, 0)))
    lead2.update(lead)
    # hoppings in different directions to test different code paths
    lead[(lat(0, 1), lat(1, 2))] = 1j * sigma_z
    lead[(lat(1, 0), lat(0, 1))] = -1j * sigma_z
    syst.attach_lead(lead)
    leads.add_voltage(syst, 0, phase_mat)
    new_hops = ((lat(-2, 1), lat(-1, 0)), (lat(-2, 1), lat(-1, 2)))
    check_hops(syst, new_hops, expected)
    # functional hoppings to test all code paths
    lead2[(lat(0, 1), lat(-1, 2))] = lambda i, j, t: 1j * sigma_z
    lead2[(lat(-1, 0), lat(0, 1))] = lambda i, j, t: -1j * sigma_z
    syst.attach_lead(lead2)
    leads.add_voltage(syst, 1, phase_mat)
    new_hops = ((lat(4, 1), lat(3, 0)), (lat(4, 1), lat(3, 2)))
    check_hops(syst, new_hops, expected)


@pytest.mark.parametrize('lead_maker', [make_simple_lead, make_complex_lead])
def test_free_boundary(lead_maker):

    # test invalid construction
    pytest.raises(ValueError, leads.SimpleBoundary, 0)
    pytest.raises(ValueError, leads.SimpleBoundary, -1)
    pytest.raises(ValueError, leads.SimpleBoundary, num_cells=0)
    pytest.raises(ValueError, leads.SimpleBoundary, num_cells=-1)
    pytest.raises(ValueError, leads.SimpleBoundary, tmax=0)
    pytest.raises(ValueError, leads.SimpleBoundary, tmax=-1)
    pytest.raises(TypeError, leads.SimpleBoundary)
    pytest.raises(TypeError, leads.SimpleBoundary, num_cells=2, tmax=2)

    ncells = 10
    lead = lead_maker(kwant.lattice.square(norbs=1)).finalized()
    norbs = lead.cell_size  # 1 orbital per site
    norbs_iface = lead.graph.num_nodes - lead.cell_size  # 1 orbital per site
    # ord=2 gives us the spectral norm
    tmax = ncells / (2 * np.linalg.norm(lead.inter_cell_hopping(), ord=2))
    eps = 1.1  # 1E-8  # boundary valid up to time = tmax + 1 instead tmax

    bdies = [((ncells,), {}), ((), dict(num_cells=ncells)),
             ((), dict(tmax=tmax))]

    for bdy in (leads.SimpleBoundary(*args, **kwargs) for args, kwargs in bdies):
        evald_bdy = bdy(lead)
        # check format of return arguments
        assert evald_bdy.from_slices == [ta.array([0, norbs])]
        assert evald_bdy.to_slices == [ta.array([0, norbs])]
        assert evald_bdy.solution_is_valid(None)
        assert evald_bdy.time_is_valid(tmax - eps)
        assert not evald_bdy.time_is_valid(tmax + eps)
        # check produced Hamiltonian
        check_boundary_hamiltonian(
            evald_bdy.hamiltonian, norbs, norbs_iface,
            lambda i: lead.cell_hamiltonian(),
            lambda i: lead.inter_cell_hopping())


@pytest.mark.parametrize('nbuffer_cells', [-1, 0, 10])
def test_monomial_absorbing_boundary(nbuffer_cells):
    ncells = 10
    strength = 10
    degree = 6
    lead = make_simple_lead(kwant.lattice.square(norbs=1)).finalized()
    norbs = lead.cell_size  # 1 orbital per site
    norbs_iface = lead.graph.num_nodes - lead.cell_size  # 1 orbital per site

    pytest.raises(ValueError, leads.MonomialAbsorbingBoundary,
                  0, strength, degree, num_buffer_cells=0)
    pytest.raises(ValueError, leads.MonomialAbsorbingBoundary,
                  ncells, strength, -1, num_buffer_cells=0)

    if nbuffer_cells < 0:
        pytest.raises(ValueError, leads.MonomialAbsorbingBoundary,
                      ncells, strength, degree, num_buffer_cells=nbuffer_cells)
        return
    else:
        bdy = leads.MonomialAbsorbingBoundary(ncells, strength, degree,
                                              num_buffer_cells=nbuffer_cells)

    evald_bdy = bdy(lead)

    # check format of return arguments
    assert evald_bdy.from_slices == [ta.array([0, norbs])]
    assert evald_bdy.to_slices == [ta.array([0, norbs])]
    assert evald_bdy.solution_is_valid(None)
    assert evald_bdy.time_is_valid(float('inf'))

    # check produced Hamiltonian

    def absorbing(i):
        if i < nbuffer_cells:
            return 0
        else:
            i = i - nbuffer_cells
            n = degree
            return -1j * np.eye(norbs) * ((n + 1) * strength *
                                          i**n / ncells**(n + 1))

    check_boundary_hamiltonian(
        evald_bdy.hamiltonian, norbs, norbs_iface,
        lambda i: lead.cell_hamiltonian() + absorbing(i),
        lambda i: lead.inter_cell_hopping())

# ---


def test_generic_absorbing_boundary():
    # test that hamiltonian is the same, if we construct our generic
    # boundary condition with the same monomial function that is
    # implemented in MonomialAbsorbingBoundary
    lead = make_simple_lead(kwant.lattice.square(norbs=1)).finalized()
    num_cells, buffer_cells, strength, degree = 100, 40, 14.5, 6

    sigma = leads._monomial_absorbing_potential(strength, degree)
    boundary_1 = leads.GenericAbsorbingBoundary(num_cells, sigma, buffer_cells)
    boundary_2 = leads.MonomialAbsorbingBoundary(num_cells, strength, degree, buffer_cells)

    h1 = boundary_1(lead).hamiltonian
    h2 = boundary_2(lead).hamiltonian
    assert np.allclose(h1.A, h2.A)


def test_absorbing_reflection_solver():
    # create explicitly a system with an imaginary monomial potential
    # and check that this gives the same reflection as by the reflection
    # solver that takes only a lead

    def create_system(L, W=1, a=1):
        gamma = 1 / a**2

        def aborbing_potential(site, strength, degree):
            x = site.pos[0]
            return - 1j / L * (degree + 1) * strength * (x / L)**degree

        def onsite(site, strength, degree):
            return 2 * gamma + aborbing_potential(site, strength, degree)

        lat = kwant.lattice.square(a=a, norbs=1)
        syst = kwant.Builder()
        syst[(lat(x, y) for x in range(L) for y in range(W))] = onsite
        syst[lat.neighbors()] = -gamma
        lead = kwant.Builder(kwant.TranslationalSymmetry((-a, 0)))
        lead[(lat(0, y) for y in range(W))] = 2 * gamma
        lead[lat.neighbors()] = -gamma
        syst.attach_lead(lead)
        return syst

    L = 30
    strength = 10
    degree = 4

    syst = create_system(L, W=2).finalized()
    band = kwant.physics.Bands(syst.leads[0])

    # introduce some small epsilon to avoid energies close to band openings
    eps = 0.0001
    energies = band(0)[0] + np.logspace(np.log10(eps), np.log10(2 - eps), 48)

    # take a lead, add the imaginary potential and calculate reflection
    # note that the system syst is not involved in this calculation but only the lead
    sigma = leads._monomial_absorbing_potential(strength, degree)

    reflection_solver = leads.AbsorbingReflectionSolver(syst.leads[0], L, sigma)
    refl = [reflection_solver(energy)[0][0] for energy in energies]

    # take the system with an imaginary potential and calculate reflection
    # note that the lead is not involved in this calculation but only syst
    def reflection(syst, E, A, n):
        return np.sqrt(kwant.smatrix(syst, E, check_hermiticity=False,
                                     params={'strength': A, 'degree': n}).transmission(0, 0))
    refl_ref = [reflection(syst, energy, strength, degree) for energy in energies]

    assert_array_almost_equal(refl, refl_ref)

    # check exceptions
    lead = syst.leads[0]
    not_a_lead = []
    not_callable = 1
    raises(TypeError, leads.AbsorbingReflectionSolver, not_a_lead, L, sigma)
    raises(ValueError, leads.AbsorbingReflectionSolver, lead, - L, sigma)
    raises(ValueError, leads.AbsorbingReflectionSolver, lead, L, not_callable)


def test_reflection_calculation():

    lead = make_simple_lead(kwant.lattice.square(norbs=1)).finalized()
    bands = kwant.physics.Bands(lead)
    num_cells, strength, degree = 100, 14.5, 6
    imaginary_potential = leads._monomial_absorbing_potential(strength, degree)

    not_callable = 1
    not_a_lead = []
    non_existing_band = 5  # band index too high

    # -------------------------------------------------------------------
    # --- reflection from energy with AbsorbingReflectionSolver
    # -------------------------------------------------------------------
    refl_solver = leads.AbsorbingReflectionSolver(lead, num_cells, imaginary_potential)

    # check for scalar energy value
    momenta, mode = 2., 0
    energies, velocities = bands(momenta, derivative_order=1)
    energies = energies[mode]
    velocities = velocities[mode]

    refl, k, vel = refl_solver(energies)
    assert_array_almost_equal(refl, [0.70653482])  # precalculated value
    assert_array_almost_equal(k, momenta)
    assert_array_almost_equal(vel, velocities)
    # check for array of energies
    energies = [0, 2.5]
    refl, k, vel = refl_solver(energies)
    # check reflection
    refl0 = [[0.59570332], [0.53411935, 0.56524117]]  # precalculated value
    [assert_array_almost_equal(i, i0) for i, i0 in zip(refl, refl0)]
    # check momenta
    k0 = lead.modes(energy=energies[0])[0].momenta[1]
    k1 = lead.modes(energy=energies[1])[0].momenta[2:]
    momenta = [k0, k1]
    [assert_array_almost_equal(i, i0) for i, i0 in zip(k, momenta)]
    # check velocities
    v0 = lead.modes(energy=energies[0])[0].velocities[1]
    v1 = lead.modes(energy=energies[1])[0].velocities[2:]
    velocities = [v0, v1]
    [assert_array_almost_equal(i, i0) for i, i0 in zip(vel, velocities)]

    # check for scalar energy value where no modes are open
    refl, k, vel = refl_solver(energies=10)
    assert refl.size == 0
    assert k.size == 0
    assert vel.size == 0

    # check for an array of energy value where only parts of the modes are open
    energies = [-5, 0.5, 3, 4.5, 11.0]
    refl, k, vel = refl_solver(energies)
    number_pos_modes = [len(lead.modes(energy=e)[0].momenta) // 2 for e in energies]
    assert np.array([r.size == i for r, i in zip(refl, number_pos_modes)]).all()
    assert np.array([r.size == i for r, i in zip(k, number_pos_modes)]).all()
    assert np.array([r.size == i for r, i in zip(vel, number_pos_modes)]).all()

    # check exeptions
    raises(TypeError, leads.AbsorbingReflectionSolver, not_a_lead, num_cells, imaginary_potential)
    raises(ValueError, leads.AbsorbingReflectionSolver, lead, - num_cells, imaginary_potential)
    raises(ValueError, leads.AbsorbingReflectionSolver, lead, num_cells, not_callable)

    # -------------------------------------------------------------------
    # --- reflection from momentum with AnalyzeReflection
    # -------------------------------------------------------------------
    analyze_reflection = leads.AnalyzeReflection(lead, num_cells, imaginary_potential)
    band = 1

    # check for scalar momentum value
    k = 1
    refl, ener, vel = analyze_reflection(k, band)
    energies, velocities = bands(k, derivative_order=1)
    assert_array_almost_equal(refl, [0.80014401727])  # precalculated value
    assert_array_almost_equal(ener, energies[band])
    assert_array_almost_equal(vel, velocities[band])

    # check for an array of momentum value
    k = [-0.2, 2]
    refl, ener, vel = analyze_reflection(k, band)
    energies = [bands(kk)[band] for kk in k]
    velocities = np.abs([bands(kk, derivative_order=1)[1][band] for kk in k])
    assert_array_almost_equal(refl, [0.56059964, 0.64908466])  # precalculated value
    assert_array_almost_equal(ener, energies)
    assert_array_almost_equal(vel, velocities)

    # calc reflection around minimum
    kmin, kmax = 0.5, 2
    refl, ener, vel, k, e0, k0 = analyze_reflection.around_extremum(kmin, kmax, band, nq=2)
    assert_array_almost_equal(refl, [0.74827915, 0.99892086, 0.99892086, 0.64908466])  # precalculated value
    assert_array_almost_equal(ener, [bands(kk)[band] for kk in k])
    assert_array_almost_equal(vel, np.abs([bands(kk, derivative_order=1)[1][band] for kk in k]))
    assert_almost_equal(e0, bands(k0)[band])

    # check exeptions
    raises(TypeError, leads.AnalyzeReflection, not_a_lead, num_cells, imaginary_potential)
    raises(ValueError, leads.AnalyzeReflection, lead, - num_cells, imaginary_potential)
    raises(ValueError, leads.AnalyzeReflection, lead, num_cells, not_callable)

    raises(AssertionError, analyze_reflection, k, non_existing_band)  # band does not exists

    raises(ValueError, analyze_reflection.around_extremum, 1, 2, band, nq=2)  # no minima inside [1,2]
    raises(AssertionError, analyze_reflection.around_extremum, kmax, kmin, band, nq=2)  # left/right swapped
    raises(AssertionError, analyze_reflection.around_extremum, 0.5, 2, non_existing_band, nq=2)  # band does not exists
    raises(ValueError, analyze_reflection.around_extremum, kmin, kmax, band, nq=-2)  # neg. nb of points
    raises(ValueError, analyze_reflection.around_extremum, kmin, kmax, band, nq=0)  # no points

    # -------------------------------------------------------------------
    # --- reflection from momentum with AnalyzeReflectionMonomial
    # -------------------------------------------------------------------
    analyze_reflection = leads.AnalyzeReflectionMonomial(lead, num_cells, strength, degree)

    # calc reflection around minimum
    kmin, kmax = 0.5, 2
    refl, ener, vel, k, e0, k0 = analyze_reflection.around_extremum(kmin, kmax, band, nq=2)

    assert_array_almost_equal(refl, [4.617257e-08, 1.000000e+00, 1.000000e+00, 7.045444e-05])  # precalculated value
    assert_array_almost_equal(ener, [bands(kk)[band] for kk in k])
    assert_array_almost_equal(vel, [bands(kk, derivative_order=1)[1][band] for kk in k])
    assert_almost_equal(e0, bands(k0)[band])

    # check exeptions
#    raises(TypeError, leads.AnalyzeReflectionMonomial, not_a_lead, num_cells, strength, degree) # TODO
    raises(ValueError, leads.AnalyzeReflectionMonomial, lead, - num_cells, strength, degree)

    raises(ValueError, analyze_reflection.around_extremum, 1, 2, band, nq=2)  # no minima inside
    raises(AssertionError, analyze_reflection.around_extremum, kmax, kmin, band, nq=2)  # left/right swapped
    raises(AssertionError, analyze_reflection.around_extremum, 0.5, 2, non_existing_band, nq=2)  # band does not exists
    raises(ValueError, analyze_reflection.around_extremum, kmin, kmax, band, nq=-2)  # neg. nb of points
    raises(ValueError, analyze_reflection.around_extremum, kmin, kmax, band, nq=0)  # no points

    # TODO: velocity sign not consistent
    # TODO: call signature of Reflection routines not consistent
    # TODO: if energies are empty, we get an error


def test_automatic_boundary():
    # for short times, simple boundary condition is more efficient
    # whereas for long times, the algorithm should select absorbing boundaries

    lead = make_simple_lead(kwant.lattice.square(norbs=1)).finalized()

    # make boundary for a single lead
    boundary = leads.automatic_boundary(lead, tmax=10)[0]
    assert type(boundary) == leads.SimpleBoundary
    boundary = leads.automatic_boundary(lead, tmax=100000)[0]
    assert type(boundary) == leads.MonomialAbsorbingBoundary

    # make boundaries for a list of leads
    _leads = [lead] * 3
    boundaries = leads.automatic_boundary(_leads, tmax=10)
    for boundary in boundaries:
        assert type(boundary) == leads.SimpleBoundary
    boundaries = leads.automatic_boundary(_leads, tmax=100000)
    for boundary in boundaries:
        assert type(boundary) == leads.MonomialAbsorbingBoundary

    # check exceptions
    not_a_lead = []
    tmax = 100000
    raises(TypeError, leads.automatic_boundary, lead=not_a_lead, tmax=tmax)  # not a lead
    raises(ValueError, leads.automatic_boundary, lead, tmax=-tmax)  # negative time
    raises(AssertionError, leads.automatic_boundary, tmax, lead)  # interchanged
    raises(ValueError, leads.automatic_boundary, lead, tmax, refl_max=-0.1)  # negative reflection
    raises(ValueError, leads.automatic_boundary, lead, tmax, degree=-1)  # degree not positive
    raises(ValueError, leads.automatic_boundary, lead, tmax, emin=0.4, emax=0.2)  # emin/emax swapped


def test_automatic_boundary_for_lead_with_flat_band():
    # for short times, simple boundary condition is more efficient
    # whereas for long times, the algorithm should select absorbing boundaries

    # edgecase with tree bands, where the second band is flat but with tiny noise
    def make_lead_with_flat_band(ll=3):
        lat = kwant.lattice.square(norbs=1)
        lead = kwant.Builder(kwant.TranslationalSymmetry((-1, 1)))
        lead[[lat(0, j) for j in range(ll)]] = 0
        lead[lat.neighbors()] = -1
        return lead

    lead = make_lead_with_flat_band().finalized()

    # make simple boundary
    boundary = leads.automatic_boundary(lead, tmax=10)[0]
    assert type(boundary) == leads.SimpleBoundary

    # make absorbing boundary
    boundary = leads.automatic_boundary(lead, tmax=100)[0]
    assert type(boundary) == leads.MonomialAbsorbingBoundary



# -*- coding: utf-8 -*-
# Copyright 2016, 2017 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

import pytest
import numpy as np
import itertools as it
import functools
import scipy.sparse as sp
import tinyarray as ta
from numpy.testing import assert_array_almost_equal, assert_almost_equal
import kwant

from .. import kernels
from ... import _common

no_sparse_blas = pytest.mark.skipif(not hasattr(kernels, 'SparseBlas'),
                                    reason='No sparse BLAS available.')


def random_vector(N):
    return np.random.rand(N) + 1j * np.random.rand(N)


@pytest.mark.parametrize(('kernel_type',), [
    (kernels.Scipy,),
    (kernels.Simple,),
    pytest.param(getattr(kernels, 'SparseBlas', None), marks=no_sparse_blas)
])
def test_kernel(kernel_type):
    N = 1000

    H_simple = sp.eye(N, dtype=complex, format='csr')
    H_rand = sp.rand(N, N, format='csr') + 1j * sp.rand(N, N, format='csr')

    class W_func:
        def __init__(self, size, w):
            self.size = size
            self.w = w

        def apply(self, time, psi, out=None):
            result = self.w.dot(psi)
            if out is not None:
                out = np.asarray(out)
                out[:] += result
            else:
                return result

    for psi_st in (None, np.zeros((N,), dtype=complex), random_vector(N)):
        for H, W in it.product((H_simple, H_rand), repeat=2):

            w_func = W_func(N, W)

            psi = random_vector(N)
            nonstatic = psi if psi_st is None else psi + psi_st
            should_be = -1j * (H.dot(psi) + W.dot(nonstatic))

            py_kern = kernel_type(H, w_func, (), psi_st=psi_st)
            c_kern = kernels.extract_c_kernel(py_kern)
            for kern in (py_kern, c_kern):
                dpsidt = random_vector(N)  # this will be overwritten anyway
                nevals = kern.nevals
                kern.rhs(psi, dpsidt, 0)
                assert kern.nevals == nevals + 1
                assert np.allclose(should_be, dpsidt)

            # now test with psi_st smaller than the total size; this will be
            # the case when we have a system with added boundary conditions.

            central_norbs = N // 2
            W_part = W.tolil()[:central_norbs, :central_norbs].tocsr()
            if psi_st is not None:
                psi_st_part = psi_st[:central_norbs]
            else:
                psi_st_part = None

            w_func = W_func(central_norbs, W_part)

            psi = random_vector(N)
            central_psi = psi[:central_norbs]
            should_be = H.dot(psi)
            nonstatic = (central_psi if psi_st is None else
                         central_psi + psi_st_part)
            should_be[:central_norbs] += W_part.dot(nonstatic)
            should_be *= -1j

            py_kern = kernel_type(H, w_func, (), psi_st=psi_st_part)
            c_kern = kernels.extract_c_kernel(py_kern)
            for kern in (py_kern, c_kern):
                dpsidt = random_vector(N)  # this will be overwritten anyway
                kern.rhs(psi, dpsidt, 0)
                assert np.allclose(should_be, dpsidt)


def test_perturbation_extractor():

    def td_hopping(i, j, time):
        ab = ta.array(sorted((i.tag, j.tag)))
        return -1 + time * kwant.digest.uniform(ab)

    lat = kwant.lattice.chain(norbs=1)
    N = 4

    syst = kwant.Builder()
    syst[(lat(i) for i in range(N))] = 1
    syst[lat.neighbors()] = td_hopping

    W = kernels.PerturbationExtractor(syst.finalized(), time_name='time', time_start=0)
    time = 3

    assert W.size == N
    assert isinstance(W.size, int)

    assert W.nnz == 6
    assert isinstance(W.nnz, int)

    wdat = W.data(time)
    assert len(wdat) == W.nnz
    assert isinstance(wdat, np.ndarray)
    assert _common.is_type_array(wdat, 'complex').all()

    row_col = W.row_col()
    assert len(row_col) == 2
    assert len(row_col[0]) == W.nnz
    assert len(row_col[1]) == W.nnz
    assert isinstance(row_col, tuple)
    assert _common.is_type_array(row_col, 'integer').all()

    empty = W.empty()
    assert empty.shape == (N, N)
    assert isinstance(empty, sp.coo_matrix)

    wt = W.evaluate(time)
    assert wt.shape == (N, N)
    assert isinstance(wt, sp.coo_matrix)
    assert np.all(wt.data == wdat)

    ket = np.random.rand(N) + 1j * np.random.rand(N)
    out = W.apply(time, ket)
    assert isinstance(out, np.ndarray)
    assert len(out) == N
    assert _common.is_type_array(out, 'complex').all()

    W = kernels.PerturbationInterpolator(syst.finalized(), time_name='time', time_start=0)

    assert W.size == N
    assert isinstance(W.size, int)

    assert W.nnz == 6
    assert isinstance(W.nnz, int)

    wt = W.evaluate(time)
    assert wt.shape == (N, N)
    assert isinstance(wt, sp.coo_matrix)
    assert np.all(wt.data == wdat)

    ket = np.random.rand(N) + 1j * np.random.rand(N)
    out = W.apply(time, ket)
    assert isinstance(out, np.ndarray)
    assert len(out) == N
    assert _common.is_type_array(out, 'complex').all()


def test_extract_perturbation():

    perturb_types = [kernels.PerturbationExtractor,
                     kernels.PerturbationInterpolator,
                     functools.partial(kernels.PerturbationInterpolator, dt=0)]

    def inner_test(fsyst, N):
        for i, perturb in enumerate(perturb_types):
            W = perturb(fsyst, time_name='time', time_start=0)
            H0 = fsyst.hamiltonian_submatrix(params={'time': 0})
            H1 = fsyst.hamiltonian_submatrix(params={'time': 1})
            ket = np.random.rand(N) + 1j * np.random.rand(N)
            # calulate W(t)
            # check return and in-place modification
            if i == 1:  # numerical tolerance for interpolated W(t)
                assert_array_almost_equal(W.evaluate(1) - (H1 - H0), 0)
            else:
                assert np.all(W.evaluate(1) == (H1 - H0))

            wmat = W.evaluate(1)  # get matrix shape
            for i in range(len(wmat.data)):
                wmat.data[i] = 0
            W.evaluate(1, wmat)    # in-place
            if i == 1:  # numerical tolerance for interpolated W(t)
                assert_array_almost_equal(wmat - (H1 - H0), 0)
            else:
                assert np.all(wmat == (H1 - H0))
            # check with wrong size for in-place modification
            pytest.raises(AssertionError, W.evaluate, 1, sp.coo_matrix((N - 1, N - 1), dtype=np.complex))
            pytest.raises(AssertionError, W.evaluate, 1, sp.coo_matrix((N + 1, N + 1), dtype=np.complex))
            # data array attribute of coo matrix has wrong size
            pytest.raises(AssertionError, W.evaluate, 1, sp.coo_matrix((N, N), dtype=np.complex))

            assert isinstance(W.evaluate(1), sp.coo_matrix)

            # calulate W(t) @ ket
            # calculations not done in same order, only check if close
            # check return and in-place modification
            assert isinstance(W.apply(1, ket), np.ndarray)
            assert np.allclose(W.apply(1, ket), np.dot(H1 - H0, ket))  # return
            out = np.zeros(N, dtype=np.complex)
            W.apply(1, ket, out)  # in-place
            assert np.allclose(out, np.dot(H1 - H0, ket))
            # check with wrong size for in-place modification
            pytest.raises(AssertionError, W.apply, 1, np.zeros((N - 1), dtype=np.complex))
            pytest.raises(AssertionError, W.apply, 1, np.zeros((N + 1), dtype=np.complex))

    def td_onsite(i, time):
        return 2 + time * kwant.digest.uniform(i.tag)

    def td_hopping(i, j, time):
        ab = ta.array(sorted((i.tag, j.tag)))
        return -1 + time * kwant.digest.uniform(ab)

    lat = kwant.lattice.chain(norbs=1)
    N = 4

    # test whole system time-dependent
    syst = kwant.Builder()
    syst[(lat(i) for i in range(N))] = td_onsite
    syst[lat.neighbors()] = td_hopping
    inner_test(syst.finalized(), N)
    # check that error is raised if norbs not given
    lat.norbs = None
    for perturb in perturb_types:
        pytest.raises(RuntimeError, perturb,
                      syst.finalized(), time_name='time', time_start=0)
    lat.norbs = 1

    # test only a partial amount time-dependent
    syst = kwant.Builder()
    syst[(lat(i) for i in range(N))] = td_onsite
    syst[(lat(i) for i in range(0, N, 2))] = 1.
    syst[lat.neighbors()] = -1
    inner_test(syst.finalized(), N)

    # test >1 orbital per site
    sigma_0 = np.eye(2)
    sigma_y = np.array([[0, -1j], [1j, 0]])

    def td_onsite2(i, time):
        U = kwant.digest.uniform(i.tag)
        return 2 + time * U * sigma_y

    def td_hopping2(i, j, time):
        U = kwant.digest.uniform(ta.array(sorted((i.tag, j.tag))))
        return -1 + time * U * sigma_0

    def td_interhopping(i, j, time):
        # *from* 2 orbital lattice *to* 1 orbital lattice
        U = kwant.digest.uniform(ta.array(sorted((i.tag, j.tag))))
        return 1j * time * U * np.array([[1, 1]])

    lat2 = kwant.lattice.chain(name='b', norbs=2)

    syst = kwant.Builder()
    syst[(lat(i) for i in range(N))] = td_onsite
    syst[(lat2(i) for i in range(N))] = td_onsite2
    syst[lat.neighbors()] = td_hopping
    syst[lat2.neighbors()] = td_hopping2
    syst[kwant.builder.HoppingKind((0,), lat, lat2)] = td_interhopping
    inner_test(syst.finalized(), 2 * N + N)


def test_perturbation_interpolation():

    def td_onsite(i, time):
        return 2 + np.sin(time * 0.1) * kwant.digest.uniform(i.tag)

    def td_hopping(i, j, time):
        ab = ta.array(sorted((i.tag, j.tag)))
        return -1 + np.sin(time * 0.13) * kwant.digest.uniform(ab)

    def td_onsite2(i, time):
        U = kwant.digest.uniform(i.tag)
        return 2 + np.cos(time * 0.12) * U * sigma_y

    def td_hopping2(i, j, time):
        U = kwant.digest.uniform(ta.array(sorted((i.tag, j.tag))))
        return -1 + np.cos(time * 0.15) * U * sigma_0

    def td_interhopping(i, j, time):
        # *from* 2 orbital lattice *to* 1 orbital lattice
        U = kwant.digest.uniform(ta.array(sorted((i.tag, j.tag))))
        return 1j * np.sin(time * 0.05) * U * np.array([[1, 1]])

    # test >1 orbital per site
    sigma_0 = np.eye(2)
    sigma_y = np.array([[0, -1j], [1j, 0]])

    N = 4
    lat = kwant.lattice.chain(norbs=1)
    lat2 = kwant.lattice.chain(name='b', norbs=2)

    syst = kwant.Builder()
    syst[(lat(i) for i in range(N))] = td_onsite
    syst[(lat2(i) for i in range(N))] = td_onsite2
    syst[lat.neighbors()] = td_hopping
    syst[lat2.neighbors()] = td_hopping2
    syst[kwant.builder.HoppingKind((0,), lat, lat2)] = td_interhopping

    fsyst = syst.finalized()

    dt = 2.0

    w = kernels.PerturbationExtractor(fsyst, time_name='time', time_start=0)
    wi = kernels.PerturbationInterpolator(fsyst, time_name='time', time_start=0, dt=dt)

    for time in [0.4, 0.5, 0.7, 1, 1.6, 2, 2.1, 8.5, 19.8]:
        time *= dt
        wt = w.evaluate(time)
        wti = wi.evaluate(time)
        assert_almost_equal(np.max(np.abs(wt - wti)), 0, decimal=5)



# Copyright 2016-2019 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.
"""Utility routines."""


import subprocess
import os
import inspect
import numpy as np

__all__ = ['version', 'TkwantDeprecationWarning', 'is_type', 'is_type_array',
           'is_not_empty', 'is_zero', 'time_start', 'time_name']

# tkwant's default initial time and time argument name
time_start = 0
time_name = 'time'

package_root = os.path.dirname(os.path.realpath(__file__))
distr_root = os.path.dirname(package_root)
version_file = '_tkwant_version.py'


def get_version_from_git():
    try:
        p = subprocess.Popen(['git', 'rev-parse', '--show-toplevel'],
                             cwd=distr_root,
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except OSError:
        return
    if p.wait() != 0:
        return
    if not os.path.samefile(p.communicate()[0].decode().rstrip('\n'), distr_root):
        # The top-level directory of the current Git repository is not the same
        # as the root directory of the tkwant distribution: do not extract the
        # version from Git.
        return

    # git describe --first-parent does not take into account tags from branches
    # that were merged-in.
    for opts in [['--first-parent'], []]:
        try:
            p = subprocess.Popen(['git', 'describe', '--long'] + opts,
                                 cwd=distr_root,
                                 stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        except OSError:
            return
        if p.wait() == 0:
            break
    else:
        return
    description = p.communicate()[0].decode().strip('v').rstrip('\n')

    release, dev, git = description.rsplit('-', 2)
    version = [release]
    labels = []
    if dev != "0":
        version.append(".dev{}".format(dev))
        labels.append(git)

    try:
        p = subprocess.Popen(['git', 'diff', '--quiet'], cwd=distr_root)
    except OSError:
        labels.append('confused')  # This should never happen.
    else:
        if p.wait() == 1:
            labels.append('dirty')

    if labels:
        version.append('+')
        version.append(".".join(labels))

    return "".join(version)


# populate the version_info dictionary with values stored in the version file
version_info = {}
with open(os.path.join(package_root, version_file), 'r') as f:
    exec(f.read(), {}, version_info)
version = version_info['version']
version_is_from_git = (version == "__use_git__")
if version_is_from_git:
    version = get_version_from_git()
    if not version:
        version = "unknown"


class TkwantDeprecationWarning(Warning):
    """Class of warnings about a deprecated feature of tkwant.

    DeprecationWarning has been made invisible by default in Python 2.7 in order
    to not confuse non-developer users with warnings that are not relevant to
    them.  In the case of tkwant, by far most users are developers, so we feel
    that a TkwantDeprecationWarning that is visible by default is useful.
    """
    pass


def ensure_isinstance(obj, typ, msg=None):
    if isinstance(obj, typ):
        return
    if msg is None:
        msg = "Expecting an instance of {}.".format(typ.__name__)
    raise TypeError(msg)

# type checking and small helper routines


def is_type(variable, generic_type, require_finite=False):
    """Return true if type(variable) matches the generic type."""
    # TODO: require finite for all types
    if generic_type == 'integer':
        return np.issubdtype(type(variable), np.integer)
    if generic_type == 'number':
        return np.issubdtype(type(variable), np.number)
    if generic_type == 'real_number':
        _type = type(variable)
        is_number = np.issubdtype(_type, np.number)
        is_complex = np.issubdtype(_type, np.complexfloating)
        if is_number and require_finite:
            is_finite = np.isfinite(variable)
            return is_number and is_finite and not is_complex
        return is_number and not is_complex
    if generic_type == 'complex':
        _type = type(variable)
        is_number = np.issubdtype(_type, np.number)
        is_complex = np.issubdtype(_type, np.complexfloating)
        return is_number and is_complex
    raise NotImplementedError('generic_type= {} not implemented'
                              .format(generic_type))


def is_type_array(array, generic_type):
    """Return true everywhere where type(array) matches the generic type"""
    array = np.array(array)
    return np.array([is_type(x, generic_type) for x in array.flatten()])


def is_zero(x, tol=1E-14):
    """Return true if |x| < tol."""
    return np.abs(x) < tol


def is_not_empty(obj):
    """Check if obj exists."""
    try:
        return obj.any() or len(obj) > 0
    except AttributeError:
        if obj:
            return True
        return False


def get_default_function_argument(func, kwarg):
    """Return the default value of a function argument.

    Parameters
    ----------
    func : callable
        Function to inspect.
    kwarg : str
        Function argument for which we like to obtain the default value.

    Returns
    -------
    default : obj
        Default argument of function `func`.
        If func(kwarg=42), the returned value would be 42.

    Notes
    -----
    A TypeError is raised, if the function `func` has no `kwarg` argument
    with a default value.
    """
    P = inspect.Parameter
    pars = inspect.signature(func).parameters  # an *ordered mapping*
    for k, v in pars.items():
        if k == kwarg and v.kind in (P.POSITIONAL_OR_KEYWORD, P.KEYWORD_ONLY):
            if v.default is not inspect._empty:
                return v.default
    try:
        func_name = func.__name__
    except AttributeError:
        func_name = '<unknown func>'   # for functools.partial and friends
    raise ValueError("Function argument {} not present in {}".
                     format(kwarg, func_name))



# -*- coding: utf-8 -*-
# Copyright 2016-2019 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.
"""Tools parallelizing tkwant with the Message Passing Interface (MPI)."""

import numpy as np
from . import _logging

__all__ = ['communicator_init', 'communicator_free', 'get_communicator',
           'distribute_dict', 'DistributedDict', 'round_robin', 'get_rank']


def get_rank():
    """Return the mpi rank of tkwants global communicator as a dict value"""
    rank = get_communicator().rank
    return {'rank': rank}


# set module logger
logger = _logging.make_logger(name=__name__)


_COMM = None  # the global MPI communicator used by tkwant by default.

# it is prefered to use ``get_communicator`` to use the
# global tkwant communicator as this will properly initialize it
# if required. alternatively one may use ``communicator_init``
# if a custom MPI communicator should be used for tkwant.


def communicator_init(comm=None):
    """Initialize a global MPI communicator for tkwant.

    Parameters
    ----------
    comm : `mpi4py.MPI.Intracomm`, optional
        MPI communicator that tkwant will use.
        By default, this routine initializes the communicator to MPI.COMM_WORLD.
        Note that this routine will always initialize the global
        communicator with a dublicate.
    """
    global _COMM
    if _COMM is None:
        if comm is None:
            from mpi4py import MPI
            _COMM = MPI.COMM_WORLD.Dup()
        else:
            _COMM = comm.Dup()
    else:
        logger.warning("tkwant MPI communicator cannot be reinitialized.")


def communicator_free():
    """Free the global tkwant MPI communicator"""
    global _COMM
    _COMM.Free()


def get_communicator(comm=None):
    """Return the global tkwant MPI communicator if input `comm` is None"""
    if comm is None:
        if _COMM is None:
            communicator_init()
        comm = _COMM
    return comm


def round_robin(comm, key):
    """Schedule integer keys in a round robin fashion to the MPI ranks.

    comm : `mpi4py.MPI.Intracomm`
        The MPI communicator.
    key : int
        Integer key.

    Returns
    -------
    run : bool
        ``run`` is true if rank == key + m * size , where m is an arbitrary
        integer, and size and rank correspond to the MPI size and rank.
    """
    return comm.rank == key % comm.size


def distribute_dict(input_dict, distribute_keys, comm):
    """Distribute a dict over all MPI ranks.

    input_dict : dict
        A dictionary that is identical on each MPI rank.
    distribute_keys : callable
        Boolean function with signature `(key, comm)` that maps each key
        from ``input_dict`` exactly to one MPI rank.
    comm : `mpi4py.MPI.Intracomm`
        The MPI communicator.

    Returns
    -------
    distributed_dict : dict
        A dict which distributed over all MPI ranks.
    """
    return {key: value for key, value in input_dict.items()
            if distribute_keys(comm, key)}


def _get_data_per_rank(data, comm, rank=None):
    """Get the data distributed over all MPI ranks.

    Notes
    -----
    If rank is None, we return a list, each list index corresponds to the
    MPI rank and the content to the data on that specific rank.
    If rank is set, we return the data on that specific rank.
    """
    if rank is None:
        data_per_rank = comm.allgather(data)
    else:
        assert 0 <= rank < comm.size
        data_per_rank = comm.bcast(data, root=rank)
    return data_per_rank


class DistributedDict():
    """A class to handle a dictionary which is distributed over MPI ranks."""

    def __init__(self, data=None, comm=None):
        """
        Initialize the dictionary.

        Parameters
        ----------
        data : dict, optional
            Dictionary with all initial one-body states.
            The dictionary can (and should be) distributed over all MPI ranks.
        comm : `mpi4py.MPI.Intracomm`, optional
            The MPI communicator over which to parallelize the computation.
            By default, use the tkwant global MPI communicator.
        """
        self.comm = get_communicator(comm)
        if data is None:
            data = {}
        self._data = data
        if not self.keys_are_unique():
            raise ValueError('Keys in input dict are dublicated')

    def add(self, key, value, rank=None, check=True):
        """Add a key value pair to the dictionary"""
        if check:
            if self.key_is_present(key):
                raise ValueError('Cannot add, key={} already present'.format(key))
        if rank is None:
            size_per_rank = self.size()
            rank = np.argmin(size_per_rank)
        if not 0 <= rank < self.comm.size:
            raise ValueError('0 <= rank < {} required, but rank={}'.
                             format(self.comm.size, rank))
        if self.comm.rank == rank:
            self._data.update({key: value})

    def delete(self, key):
        """Delete a dictionary entry"""
        if self.comm.rank == self.rank_from_key(key):
            del self._data[key]

    def data(self, key):
        """Get the data corresponding to the key"""
        data = self._data.get(key)
        rank = self.rank_from_key(key)
        return self.comm.bcast(data, root=rank)

    def local_data(self):
        """Get all data stored on the local MPI ranks"""
        return self._data

    def local_keys(self):
        """Get all keys on the local MPI ranks"""
        return list(self._data.keys())

    def keys(self, rank=None):
        """Get all keys of the dictionary"""
        my_keys = list(self._data.keys())
        keys_per_rank = _get_data_per_rank(my_keys, self.comm, rank)
        if rank is None:  # flatten data to one list
            return [y for x in keys_per_rank for y in x]
        return keys_per_rank

    def size(self, rank=None):
        """Get the size of the dictonary"""
        size = len(self._data)
        return np.array(_get_data_per_rank(size, self.comm, rank))

    def move_data(self, key, to_rank):
        """Move the data to a specific MPI rank"""
        rank = self.rank_from_key(key)
        if self.comm.rank == rank:
            self.comm.send(self._data[key], dest=to_rank, tag=44)
            del self._data[key]
        if self.comm.rank == to_rank:
            self._data[key] = self.comm.recv(source=rank, tag=44)

    def rank_from_key(self, key):
        """Get the MPI rank that hosts a specific key data"""
        for rank in range(self.comm.size):
            if key in self.keys(rank=rank):
                return rank
        raise ValueError('Cannot find key={}'.format(key))

    def key_is_present(self, key):
        """Returns true if a key is present in the dataset"""
        try:
            self.rank_from_key(key)
        except ValueError:
            return False
        return True

    def keys_are_unique(self):
        """Returns true the stored dictionary data is not dublicated on MPI ranks"""
        keys = self.keys()
        if len(set(list(keys))) == len(keys):
            return True
        return False



# -*- coding: utf-8 -*-

IF SPARSELIB == "rsb":
    # librsb requires that rsb.h always be included before "blas_sparse.h"
    cdef extern from "rsb.h":
        pass

cdef extern from "blas_sparse.h":

    enum blas_order_type:
        blas_rowmajor = 101
        blas_colmajor = 102

    enum blas_trans_type:
        blas_no_trans   = 111
        blas_trans      = 112
        blas_conj_trans = 113

    enum blas_uplo_type:
        blas_upper = 121
        blas_lower = 122

    enum blas_diag_type:
        blas_non_unit_diag = 131
        blas_unit_diag     = 132

    enum blas_side_type:
        blas_left_side  = 141
        blas_right_side = 142

    enum blas_cmach_type:
        blas_base      = 151
        blas_t         = 152
        blas_rnd       = 153
        blas_ieee      = 154
        blas_emin      = 155
        blas_emax      = 156
        blas_eps       = 157
        blas_prec      = 158
        blas_underflow = 159
        blas_overflow  = 160
        blas_sfmin     = 161

    enum blas_norm_type:
        blas_one_norm       = 171
        blas_real_one_norm  = 172
        blas_two_norm       = 173
        blas_frobenius_norm = 174
        blas_inf_norm       = 175
        blas_real_inf_norm  = 176
        blas_max_norm       = 177
        blas_real_max_norm  = 178

    enum blas_sort_type:
        blas_increasing_order = 181
        blas_decreasing_order = 182

    enum blas_conj_type:
        blas_conj    = 191
        blas_no_conj = 192

    enum blas_jrot_type:
        blas_jrot_inner  = 201
        blas_jrot_outer  = 202
        blas_jrot_sorted = 203

    enum blas_prec_type:
        blas_prec_single     = 211
        blas_prec_double     = 212
        blas_prec_indigenous = 213
        blas_prec_extra      = 214

    enum blas_base_type:
        blas_zero_base = 221
        blas_one_base  = 222

    enum blas_symmetry_type:
        blas_general          = 231
        blas_symmetric        = 232
        blas_hermitian        = 233
        blas_triangular       = 234
        blas_lower_triangular = 235
        blas_upper_triangular = 236
        blas_lower_symmetric  = 237
        blas_upper_symmetric  = 238
        blas_lower_hermitian  = 239
        blas_upper_hermitian  = 240

    enum blas_field_type:
        blas_complex          = 241
        blas_real             = 242
        blas_double_precision = 243
        blas_single_precision = 244

    enum blas_size_type:
        blas_num_rows      = 251
        blas_num_cols      = 252
        blas_num_nonzeros  = 253

    enum blas_handle_type:
        blas_invalid_handle = 261
        blas_new_handle     = 262
        blas_open_handle    = 263
        blas_valid_handle   = 264

    enum blas_sparsity_optimization_type:
        blas_regular       = 271
        blas_irregular     = 272
        blas_block         = 273
        blas_unassembled   = 274


    ctypedef int blas_sparse_matrix

    ### Level 1 Computational Routines

    cdef void BLAS_susdot( blas_conj_type conj, int nz, const float *x,
                      const int *indx, const float *y, int incy, float *r,
                      blas_base_type index_base )
    cdef void BLAS_dusdot( blas_conj_type conj, int nz, const double *x,
                      const int *indx, const double *y, int incy, double *r,
                      blas_base_type index_base )
    cdef void BLAS_cusdot( blas_conj_type conj, int nz, const void *x,
                      const int *indx, const void *y, int incy, void *r,
                      blas_base_type index_base )
    cdef void BLAS_zusdot( blas_conj_type conj, int nz, const void *x,
                      const int *indx, const void *y, int incy, void *r,
                      blas_base_type index_base )

    cdef void BLAS_susaxpy( int nz, float alpha, const float *x, const int *indx,
                     float *y, int incy, blas_base_type index_base )
    cdef void BLAS_dusaxpy( int nz, double alpha, const double *x, const int *indx,
                     double *y, int incy, blas_base_type index_base )
    cdef void BLAS_cusaxpy( int nz, const void *alpha, const void *x, const int *indx,
                     void *y, int incy, blas_base_type index_base )
    cdef void BLAS_zusaxpy( int nz, const void *alpha, const void *x, const int *indx,
                     void *y, int incy, blas_base_type index_base )

    cdef void BLAS_susga( int nz, const float *y, int incy, float *x, const int *indx,
                  blas_base_type index_base )
    cdef void BLAS_dusga( int nz, const double *y, int incy, double *x, const int *indx,
                  blas_base_type index_base )
    cdef void BLAS_cusga( int nz, const void *y, int incy, void *x, const int *indx,
                  blas_base_type index_base )
    cdef void BLAS_zusga( int nz, const void *y, int incy, void *x, const int *indx,
                  blas_base_type index_base )

    cdef void BLAS_susgz( int nz, float *y, int incy, float *x, const int *indx,
                  blas_base_type index_base )
    cdef void BLAS_dusgz( int nz, double *y, int incy, double *x, const int *indx,
                   blas_base_type index_base )
    cdef void BLAS_cusgz( int nz, void *y, int incy, void *x, const int *indx,
                   blas_base_type index_base )
    cdef void BLAS_zusgz( int nz, void *y, int incy, void *x, const int *indx,
                   blas_base_type index_base )

    cdef void BLAS_sussc( int nz, const float *x, float *y, int incy, const int *indx,
                   blas_base_type index_base )
    cdef void BLAS_dussc( int nz, const double *x, double *y, int incy, const int *indx,
                   blas_base_type index_base )
    cdef void BLAS_cussc( int nz, const void *x, void *y, int incy, const int *indx,
                   blas_base_type index_base )
    cdef void BLAS_zussc( int nz, const void *x, void *y, int incy, const int *indx,
                   blas_base_type index_base )

    ### Level 2 Computational Routines

    cdef int BLAS_susmv(  blas_trans_type transa, float alpha,
        blas_sparse_matrix A, const float *x, int incx, float *y, int incy )
    cdef int BLAS_dusmv(  blas_trans_type transa, double alpha,
        blas_sparse_matrix A, const double *x, int incx, double *y, int incy )
    cdef int BLAS_cusmv(  blas_trans_type transa, const void *alpha,
        blas_sparse_matrix A, const void *x, int incx, void *y, int incy )
    cdef int BLAS_zusmv(  blas_trans_type transa, const void *alpha,
        blas_sparse_matrix A, const void *x, int incx, void *y, int incy )

    cdef int BLAS_sussv(  blas_trans_type transt, float alpha,
        blas_sparse_matrix T, float *x, int incx )
    cdef int BLAS_dussv(  blas_trans_type transt, double alpha,
        blas_sparse_matrix T, double *x, int incx )
    cdef int BLAS_cussv(  blas_trans_type transt, const void *alpha,
        blas_sparse_matrix T, void *x, int incx )
    cdef int BLAS_zussv(  blas_trans_type transt, const void *alpha,
        blas_sparse_matrix T, void *x, int incx )

    ### Level 3 Computational Routines

    cdef int BLAS_susmm(  blas_order_type order,  blas_trans_type transa,
        int nrhs, float alpha, blas_sparse_matrix A, const float *b, int ldb,
            float *c, int ldc )
    cdef int BLAS_dusmm(  blas_order_type order,  blas_trans_type transa,
            int nrhs, double alpha, blas_sparse_matrix A, const double *b,
            int ldb, double *c, int ldc )
    cdef int BLAS_cusmm(  blas_order_type order,  blas_trans_type transa,
             int nrhs, const void *alpha, blas_sparse_matrix A, const void *b,
         int ldb, void *c, int ldc )
    cdef int BLAS_zusmm(  blas_order_type order,  blas_trans_type transa,
             int nrhs, const void *alpha, blas_sparse_matrix A, const void *b,
         int ldb, void *c, int ldc )

    cdef int BLAS_sussm(  blas_order_type order,  blas_trans_type transt,
                  int nrhs, float alpha, int t, float *b, int ldb )
    cdef int BLAS_dussm(  blas_order_type order,  blas_trans_type transt,
                  int nrhs, double alpha, int t, double *b, int ldb )
    cdef int BLAS_cussm(  blas_order_type order,  blas_trans_type transt,
                  int nrhs, const void *alpha, int t, void *b, int ldb )
    cdef int BLAS_zussm(  blas_order_type order,  blas_trans_type transt,
                  int nrhs, const void *alpha, int t, void *b, int ldb )

    ### Handle Management Routines

    ### Creation Routines

    cdef blas_sparse_matrix BLAS_suscr_begin( int m, int n )
    cdef blas_sparse_matrix BLAS_duscr_begin( int m, int n )
    cdef blas_sparse_matrix BLAS_cuscr_begin( int m, int n )
    cdef blas_sparse_matrix BLAS_zuscr_begin( int m, int n )


    cdef blas_sparse_matrix BLAS_suscr_block_begin( int Mb, int Nb, int k, int l )
    cdef blas_sparse_matrix BLAS_duscr_block_begin( int Mb, int Nb, int k, int l )
    cdef blas_sparse_matrix BLAS_cuscr_block_begin( int Mb, int Nb, int k, int l )
    cdef blas_sparse_matrix BLAS_zuscr_block_begin( int Mb, int Nb, int k, int l )

    cdef blas_sparse_matrix BLAS_suscr_variable_block_begin( int Mb, int Nb,
                    const int *k, const int *l )
    cdef blas_sparse_matrix BLAS_duscr_variable_block_begin( int Mb, int Nb,
                    const int *k, const int *l )
    cdef blas_sparse_matrix BLAS_cuscr_variable_block_begin( int Mb, int Nb,
                    const int *k, const int *l )
    cdef blas_sparse_matrix BLAS_zuscr_variable_block_begin( int Mb, int Nb,
                    const int *k, const int *l )


    ### Insertion Routines

    cdef int BLAS_suscr_insert_entry( blas_sparse_matrix A, float val, int i, int j )
    cdef int BLAS_duscr_insert_entry( blas_sparse_matrix A, double val, int i, int j )
    cdef int BLAS_cuscr_insert_entry( blas_sparse_matrix A, const void *val, int i, int j )
    cdef int BLAS_zuscr_insert_entry( blas_sparse_matrix A, const void *val, int i, int j )

    cdef int BLAS_suscr_insert_entries( blas_sparse_matrix A, int nz, const float *val,
                                const int *indx, const int *jndx )
    cdef int BLAS_duscr_insert_entries( blas_sparse_matrix A, int nz, const double *val,
                                const int *indx, const int *jndx )
    cdef int BLAS_cuscr_insert_entries( blas_sparse_matrix A, int nz, const void *val,
                                const int *indx, const int *jndx )
    cdef int BLAS_zuscr_insert_entries( blas_sparse_matrix A, int nz, const void *val,
                                const int *indx, const int *jndx )

    cdef int BLAS_suscr_insert_col( blas_sparse_matrix A, int j, int nz,
                               const float *val, const int *indx )
    cdef int BLAS_duscr_insert_col( blas_sparse_matrix A, int j, int nz,
                               const double *val, const int *indx )
    cdef int BLAS_cuscr_insert_col( blas_sparse_matrix A, int j, int nz,
                               const void *val, const int *indx )
    cdef int BLAS_zuscr_insert_col( blas_sparse_matrix A, int j, int nz,
                               const void *val, const int *indx )

    cdef int BLAS_suscr_insert_row( blas_sparse_matrix A, int i, int nz,
                               const float *val, const int *indx )
    cdef int BLAS_duscr_insert_row( blas_sparse_matrix A, int i, int nz,
                               const double *val, const int *indx )
    cdef int BLAS_cuscr_insert_row( blas_sparse_matrix A, int i, int nz,
                               const void *val, const int *indx )
    cdef int BLAS_zuscr_insert_row( blas_sparse_matrix A, int i, int nz,
                               const void *val, const int *indx )

    cdef int BLAS_suscr_insert_clique( blas_sparse_matrix A, const int k, const int l,
                            const float *val, const int row_stride,
                            const int col_stride, const int *indx,
                            const int *jndx )
    cdef int BLAS_duscr_insert_clique( blas_sparse_matrix A, const int k, const int l,
                            const double *val, const int row_stride,
                            const int col_stride, const int *indx,
                            const int *jndx )
    cdef int BLAS_cuscr_insert_clique( blas_sparse_matrix A, const int k, const int l,
                            const void *val, const int row_stride,
                            const int col_stride, const int *indx,
                            const int *jndx )
    cdef int BLAS_zuscr_insert_clique( blas_sparse_matrix A, const int k, const int l,
                            const void *val, const int row_stride,
                            const int col_stride, const int *indx,
                            const int *jndx )

    cdef int BLAS_suscr_insert_block( blas_sparse_matrix A, const float *val,
                            int row_stride, int col_stride, int i, int j )
    cdef int BLAS_duscr_insert_block( blas_sparse_matrix A, const double *val,
                            int row_stride, int col_stride, int i, int j )
    cdef int BLAS_cuscr_insert_block( blas_sparse_matrix A, const void *val,
                            int row_stride, int col_stride, int i, int j )
    cdef int BLAS_zuscr_insert_block( blas_sparse_matrix A, const void *val,
                            int row_stride, int col_stride, int i, int j )

    ### Completion of Construction Routines

    cdef int BLAS_suscr_end( blas_sparse_matrix A )
    cdef int BLAS_duscr_end( blas_sparse_matrix A )
    cdef int BLAS_cuscr_end( blas_sparse_matrix A )
    cdef int BLAS_zuscr_end( blas_sparse_matrix A )

    ### Matrix Property Routines

    cdef int BLAS_usgp( blas_sparse_matrix A, int pname )
    cdef int BLAS_ussp( blas_sparse_matrix A, int pname )

    ### Destruction Routine

    cdef int BLAS_usds( blas_sparse_matrix A )



import tkwant
from tkwant import leads, manybody
import kwant
import kwantspectrum

import functools
import cmath
import numpy as np
from scipy.special import erf
import matplotlib.pyplot as plt


def am_master():
    """Return true for the MPI master rank"""
    return tkwant.mpi.get_communicator().rank == 0


def make_system(L=400):

    def gaussian(time, t0=40, A=1.57, sigma=24):
        return A * (1 + erf((time - t0) / sigma))

    # time dependent coupling with gaussian pulse
    def coupling_nn(site1, site2, time):
        return - cmath.exp(- 1j * gaussian(time))

    # system building
    lat = kwant.lattice.square(a=1, norbs=1)
    syst = kwant.Builder()

    # central scattering region
    syst[(lat(x, 0) for x in range(L))] = 1
    syst[lat.neighbors()] = -1
    # time dependent coupling between two sites in the center
    syst[lat(L // 2, 0), lat(L // 2 - 1, 0)] = coupling_nn

    # add leads
    sym = kwant.TranslationalSymmetry((-1, 0))
    lead_left = kwant.Builder(sym)
    lead_left[lat(0, 0)] = 1
    lead_left[lat.neighbors()] = -1
    syst.attach_lead(lead_left)
    syst.attach_lead(lead_left.reversed())

    return syst


def main():

    syst = make_system().finalized()
    sites = [site.pos[0] for site in syst.sites]
    times = [40, 80, 120, 160]

    density_operator = kwant.operator.Density(syst)

    # calculate the spectrum E(k) for all leads
    spectra = kwantspectrum.spectra(syst.leads)

    # estimate the cutoff energy Ecut from T, \mu and f(E)
    # All states are effectively empty above E_cut
    occupations = manybody.lead_occupation(chemical_potential=0, temperature=0)
    emin, emax = manybody.calc_energy_cutoffs(occupations)

    # define boundary conditions
    bdr = leads.automatic_boundary(spectra, tmax=max(times), emin=emin, emax=emax)

    # calculate the k intervals for the quadrature
    interval_type = functools.partial(manybody.Interval, order=20,
                                      quadrature='gausslegendre')
    intervals = manybody.calc_intervals(spectra, occupations, interval_type)
    intervals = manybody.split_intervals(intervals, number_subintervals=10)

    # calculate all onebody scattering states at t = 0
    tasks = manybody.calc_tasks(intervals, spectra, occupations)
    psi_init = manybody.calc_initial_state(syst, tasks, bdr)

    # set up the manybody wave function
    wave_function = manybody.WaveFunction(psi_init, tasks)

    density0 = wave_function.evaluate(density_operator)

    for time in times:
        wave_function.evolve(time)
        density = wave_function.evaluate(density_operator)
        if am_master():
            plt.plot(sites, density - density0, label='time={}'.format(time))

    if am_master():
        plt.legend()
        plt.xlabel(r'site position $i$')
        plt.ylabel(r'charge density $n$')
        plt.show()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# cython: embedsignature=True
#
# Copyright 2016-2020 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.
"""Calculating the right-hand side of the time-dependent Schrödinger equation.

This module is mainly for internal use.

The classes in this module are used for calculating the following expression:

    ..math:: -i(H_0 ψ + W(t) ψ)

where :math:`H_0` is the time-independent part of the Hamiltonian,
:math:`W(t)` is the time-dependent part, and :math:`ψ` is some vector
defined over the Hilbert space of the problem. We recognise this expression
as the right-hand side of the time-dependent Schrödinger equation (TDSE),
where the left-hand side is :math:`∂ψ/∂t`.

In practice we will want to handle many different cases, including solving the
TDSE on infinite domains, or starting the evolution from an eigenstate.  All
the forms of TDSE that we will need to handle these cases are encompassed by
the above formula. This module is only concerned with calculating expressions
of the above form, not performing the mapping from the various problems. For
more details on the mapping, see `tkwant/onebody/solver.py`.
"""

__all__ = ['Kernel', 'CKernel', 'Scipy', 'Simple',
           'extract_c_kernel', 'default', 'PerturbationExtractor',
           'PerturbationInterpolator']

import numpy as np
import cython
from cpython.ref cimport Py_INCREF, Py_DECREF
from .. import _common, system
from scipy.sparse._sparsetools import coo_matvec, csr_matvec
import scipy.sparse as sp


########## Abstract Base Classes for kernels

cdef class Kernel:
    """ABC for right-hand side of the time-dependent Schrödinger equation.

    Attributes
    ----------
    size : int
        The size of the time-independent part of the Hamiltonian. This
        also sets the size of the solution vector.
    nevals : int
        The number of times this kernel has been evaluated since its creation.
    """

    def __init__(self, H0, W, params, complex[:] psi_st=None):
        raise NotImplementedError()

    def set_params(self, params):
        raise NotImplementedError()

    def rhs(self, complex[:] psi, complex[:] dpsidt, double time):
        """Evaluate the RHS of the TDSE and store the result in `dpsidt`."""
        raise NotImplementedError()


### C-level interface to Kernels (needed for C-implemented solvers)


cdef class CKernel(Kernel):
    """Base class for kernels that directly provide a C-level interface.

    Attributes
    ----------
    c_self : void*
        Pointer to a C struct that provides extra context (e.g. H0 and W)
        to the kernel.
    c_rhs : void (*)(const void* c_self, const complex* psi, complex* dpsidt,
                     double time)
        A C function that evaluates the right-hand side of the Schrödinger
        equation, and stores the result in 'dpsidt'.
    """

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def rhs(self, const complex[::1] psi, complex[::1] dpsidt, double time):
        """Evaluate the RHS of the TDSE and store the result in `dpsidt`."""
        self.c_rhs(self.c_self, &psi[0], &dpsidt[0], time)
        self.nevals += 1


# C wrapped to access pure Python Kernel implementation
cdef void _wrapped_rhs(const void *self, const complex *psi, complex *dpsidt,
                       double time) except *:
    cdef size_t size = (<object>self).size
    (<object>self).rhs(<complex[:size]>psi, <complex[:size]>dpsidt, time)


cdef class PyCKernel(CKernel):
    """C interface to python-implemented kernels."""

    def __init__(self, py_kernel):
        pass

    def __cinit__(self, py_kernel):
        self.size = py_kernel.size
        self.nevals = py_kernel.nevals
        self.c_rhs = _wrapped_rhs
        self.c_self = <void*>py_kernel
        # increment the reference count to prevent garbage collection
        Py_INCREF(<object>self.c_self)

    def __dealloc__(self):
        Py_DECREF(<object>self.c_self)


cpdef CKernel extract_c_kernel(Kernel py_kernel):
    """Extract a C-level interface from a (possibly pure-python) Kernel."""
    if isinstance(py_kernel, CKernel):
        return py_kernel
    elif not callable(py_kernel.rhs):
        raise TypeError('Kernel must have a `rhs` method.')
    return PyCKernel(py_kernel)


########## Concrete kernel implementations

class Scipy(Kernel):
    """Evaluate the RHS of the Schrödinger equation using scipy sparse.

    Parameters
    ----------
    H0 : `scipy.sparse.base`
        Hamiltonian at ``t=0`` (including any boundary conditions).
    W : callable
        Time-dependent part of the Hamiltonian. Typically the object returned
        by `tkwant.system.extract_perturbation`.
    params : dict
        Extra arguments to pass to the system Hamiltonian excluding time.
    psi_st : array of complex, optional
        The wavefunction of the initial eigenstate defined over the central
        system (if starting in an initial eigenstate).

    Attributes
    ----------
    H0 : `scipy.sparse.csr_matrix`
        Hamiltonian at ``t=0`` (including any boundary conditions).
    W : callable
        Time-dependent part of the Hamiltonian. Typically the object returned
        by `tkwant.system.extract_perturbation`.
    params : dict
        Extra arguments to pass to the system Hamiltonian excluding the ``time``
        argument.
    psi_st : array of complex or `None`
        The wavefunction of the initial eigenstate defined over the central
        system (if starting in an initial eigenstate).
    size : int
        The size of the time-independent part of the Hamiltonian. This
        also sets the size of the solution vector.
    nevals : int
        The number of times this kernel has been evaluated since its creation.
    """

    def __init__(self, H0, W, params, psi_st=None):
        self.H0 = H0.tocsr()
        self.W = W
        self.psi_st = psi_st
        self.size = self.H0.shape[0]
        if W is not None:
            self.center_size = self.W.size
        else:
            self.center_size = self.size
        self.nevals = 0
        self.params = params
        self._tmp = np.zeros((self.size,), dtype=complex)

    def set_params(self, params):
        self.params = params

    def rhs(self, const complex[:] psi, complex[:] dpsidt, double time):
        """Evaluate the RHS of the TDSE and store the result in `dpsidt`."""
        psi_centre = np.asarray(psi[:self.center_size])
        # H0 @ psi -> tmp
        cdef complex[:] tmp = self._tmp
        tmp[:] = 0
        csr_matvec(
            self.H0.shape[0], self.H0.shape[1], self.H0.indptr, self.H0.indices, self.H0.data,
            psi, tmp)
#         cdef complex[:] tmp = self.H0.dot(psi)
        # W(t) @ (psi + psi_st) + tmp -> tmp
        if self.psi_st is not None:
            psi_centre = psi_centre + self.psi_st

        try:
            self.W.apply(time, psi_centre[:self.center_size], out=tmp[:self.center_size])
        except Exception as exept:
            if self.W is None:
                pass
            else:
                raise exept

        # dpsidt = -1j * tmp
        cdef int i
        for i in range(self.size):
            dpsidt[i] = -1j * tmp[i]

        self.nevals += 1


@cython.boundscheck(False)
@cython.wraparound(False)
cdef(double, double, double, double, double) update_times(double t, double t0, double dt):
    '''Return new update times for the Hamiltonian matrix.
    '''
    assert dt > 0
    cdef int quotient
    if t < t0:
        raise ValueError('time t={} must be greater equal t0={}'.format(t, t0))
    if t >= t0 + dt + dt:
        quotient = int(round((t - t0) / dt, 10))  # integer division // has problems with the numerics
        t0 = t0 + quotient * dt
    # adding dt up we get inaccurate routings
    cdef double t1 = t0 + dt
    cdef double t2 = t0 + 2 * dt
    cdef double t3 = t0 + 3 * dt
    cdef double t4 = t0 + 4 * dt
    return t0, t1, t2, t3, t4


@cython.boundscheck(False)
@cython.wraparound(False)
cdef(double, double, double, double) _get_polynome_coeffs(double t, double t0, double t1, double dt):

    cdef double rdx = 1 / dt
    cdef double fac = dt * dt / 6.

    cdef double a = (t1 - t) * rdx
    cdef double b = (t - t0) * rdx
    cdef double c = (a * a * a - a) * fac
    cdef double d = (b * b * b - b) * fac
    return a, b, c, d


def _herm_conj(value):
    if hasattr(value, 'conjugate'):
        value = value.conjugate()
    if hasattr(value, 'transpose'):
        value = value.transpose()
    return value


_errormsg = ('{0} vector has length {1}, but the system '
             'contains {2} orbitals')


# TODO: this only works for finalized builders so far -- how to extend
#       this to something more general?
# TODO: change this routine when kwant supports vectorized systems
cdef class PerturbationExtractor:
    """Extract the time-dependent perturbation to the Hamiltonian.

    The total Hamiltonian can be split into two parts: the stationary
    part and the time-dependent perturbation, i.e.  H(t) = H_0 + W(t).
    This routine extracts the W(t) part from the Kwant system that
    specifies the H(t).

    Parameters
    ----------
    syst : `kwant.builder.FiniteSystem`
        The system from which to extract the time-dependent perturbation.
    time_name : str
        The name of the time argument. Only sites with Hamiltonian value
        functions with the argument name `time_name` are extracted into W(t).
    time_start : float
        The initial time.
    params : dict, optional
        Extra arguments to pass to the Hamiltonian of ``syst``, excluding time.

    Notes
    -----
    By definition the perturbation is defined with respect to the initial time.
    The perturbation should be therefore zero for times t < `time_start`. This
    is not inforced by this routine however, but W(t) is always evaluated
    with the time argument it gets.
    """
    cdef:
        object H
        object tparams
        int size
        int nnz
        object time_name
        object td_hamiltonian

    def __init__(self, syst, time_name, time_start, params=None):
        self.tparams = system.add_time_to_params(params, time_name, time_start)
        self.time_name = time_name
        self.H = H = syst.hamiltonian

        # Number of (potentially) non-zero entries in the time-dependent
        # Hamiltonian matrix.
        nnz = 0

        # Process onsites.
        self.td_hamiltonian = td_hamiltonian = []
        for i, (onsite, _) in enumerate(syst.onsites):
            if system.is_time_dependent_function(onsite, time_name):
                a, b = rng = system.orb_range(syst, i)
                norbs = b - a
                nnz += norbs * norbs
                td_hamiltonian.append(
                    ((i, i, rng, rng, H(i, i, params=self.tparams))))

        # Process hoppings.
        for edge_id, (i, j) in enumerate(syst.graph):
            hop_val, _ = syst.hoppings[edge_id]
            if system.is_time_dependent_function(hop_val, time_name):
                a, b = ori = system.orb_range(syst, i)
                c, d = orj = system.orb_range(syst, j)
                nnz += 2 * (b - a) * (d - c)
                td_hamiltonian.append((i, j, ori, orj, H(i, j, params=self.tparams)))

        _, _, total_norbs = syst.site_ranges[-1]
        self.nnz = nnz
        self.size = total_norbs

    @property
    def size(self):
        '''The size of the W(t) matrix.

        Returns
        -------
        size : int
            The size of the W(t) matrix is ``size`` x ``size``.
        '''
        return self.size

    @property
    def nnz(self):
        '''The total number of time-dependent matrix elements.

        Returns
        -------
        nnz : int
            The total number of time-dependent matrix elements, equal to
            len(W(t).data).
        '''
        return self.nnz

    def data(self, double time, object out=None):
        '''Return the time-dependent matrix elements.

        Parameters
        ----------
        time : int or float
            Time argument, must be equal or larger than `time_start`.
        out : `numpy.ndarray`, optional
            Data array of the W(t) matrix for operation in-place.

        Returns
        -------
        out : `numpy.ndarray` or `None`
            Data array of the W(t) matrix (W(t).data). If an ``out`` argument is given,
            the operation is performed in-place and the routine returns `None`.
        '''
        if out is None:
            do_return = True
            out = np.empty(self.nnz, complex)
        else:
            do_return = False
            assert out.shape[0] == self.nnz, _errormsg.format('output',
                                                              out.shape[0],
                                                              self.nnz)
        index = 0
        self.tparams[self.time_name] = time
        for i, j, _, _, H_0 in self.td_hamiltonian:
            mat_el = np.asarray(self.H(i, j, params=self.tparams) - H_0)
            for val in mat_el.flatten():
                out[index] = val
                index += 1
            if i != j:
                for val in _herm_conj(mat_el).flatten():
                    out[index] = val
                    index += 1
        return out if do_return else None

    def row_col(self):
        '''Return the row and column vectors of the W(t) matrix.

        Returns
        -------
        row : `np.ndarray`
            Row vector of the W(t) matrix in coo format.
        col : `np.ndarray`
            Column vector of the W(t) matrix in coo format.
        '''
        row = np.empty(self.nnz, np.int32)
        col = np.empty(self.nnz, np.int32)
        index = 0
        for i, j, to, frm, H_0 in self.td_hamiltonian:
            to = range(*to)
            frm = range(*frm)
            for k in to:
                for l in frm:
                    row[index] = k
                    col[index] = l
                    index += 1
            if i != j:
                for k in to:
                    for l in frm:
                        row[index] = l
                        col[index] = k
                        index += 1
        return row, col

    def empty(self):
        '''Return an empty W(t) matrix with correct shape and indices.

        Returns
        -------
        out : `~scipy.sparse.coo_matrix`
            An "empty" W(t) matrix filled with zeros.
        '''
        data = np.zeros((self.nnz,), dtype=np.complex)
        row_col = self.row_col()
        return sp.coo_matrix((data, row_col), (self.size, self.size))

    def evaluate(self, double time, object out=None):
        '''Evaluate the W(t) matrix for the given time t.

        Parameters
        ----------
        time : int or float
            Time argument, must be equal or larger than `time_start`.
        out : `~scipy.sparse.coo_matrix`, optional
            Sparse matrix with to perform the operation in-place.

        Returns
        -------
        out : `~scipy.sparse.coo_matrix` or `None`
            The matrix W(t).
            The size of the W(t) matrix is ``size`` x ``size``, containing
            ``nnz`` complex entries. If an ``out`` argument is given,
            the operation is performed in-place and the routine returns `None`.
        '''
        if out is None:
            data = self.data(time)
            row_col = self.row_col()
            return sp.coo_matrix((data, row_col), (self.size, self.size))
        else:
            assert out.shape[0] == self.size, _errormsg.format('output',
                                                                out.shape[0],
                                                                self.size)
            self.data(time, out.data)

    def apply(self, double time, ket, out=None):
        '''Evaluate the matrix-vector product W(t) @ ket for the given time.

        Parameters
        ----------
        time : int or float
            Time argument, must be equal or larger than `time_start`.
        out : `numpy.ndarray`, optional
            Sparse matrix with to perform the operation in-place.

        Returns
        -------
        out : `numpy.ndarray` or `None`
            The product W(t) @ ket. If an ``out`` argument is given,
            the operation is performed in-place and the routine returns `None`.
        '''
        assert ket.shape[0] == self.size, _errormsg.format('Ket', ket.shape[0],
                                                           self.size)
        if out is None:
            do_return = True
            out = np.zeros((self.size,), dtype=np.complex)
        else:
            do_return = False
            assert out.shape[0] == self.size, _errormsg.format('output',
                                                               out.shape[0],
                                                               self.size)
        self.tparams[self.time_name] = time
        for i, j, to, frm, H_0 in self.td_hamiltonian:
            to = slice(*to)
            frm = slice(*frm)
            mat_el = self.H(i, j, params=self.tparams) - H_0
            out[to] += np.dot(mat_el, ket[frm])
            if i != j:
                out[frm] += np.dot(_herm_conj(mat_el), ket[to])
        return out if do_return else None


cdef class PerturbationInterpolator:
    r"""A class to extract and interpolate W(t).

    This class can be used like `tkwant.onebody.kernel.PerturbationExtractor`,
    to calculate W(t) and the matrix-vector product W(t) * ket.
    Interpolation is done only for performance reasons, as W(t) from
    Kwant is slow to evaluate.

    Parameters
    ----------
    syst : `kwant.builder.FiniteSystem`
        The system from which to extract the time-dependent perturbation.
    time_name : str
        The name of the time argument. Only sites with Hamiltonian value
        functions with the argument name `time_name` are extracted into W(t).
    time_start : float
        The initial time.
    params : dict, optional
        Extra arguments to pass to the Hamiltonian of ``syst``, excluding time.
    dt : float
        Discretization timestep for the Spline. Must be => 0.
        For dt = 0, no interpolation is performed, but the result is
        identical to `tkwant.onebody.kernel.PerturbationExtractor`.

    Notes
    -----
    By definition the perturbation is defined with respect to the initial time.
    The perturbation should be therefore zero for times t < `time_start`. This
    is not inforced by this routine however, but W(t) is always evaluated
    with the time argument it gets.
    """

    cdef double dt
    cdef double t0, t1, t2, t3, t4
    cdef:
        object Ht0
        unsigned size
        double time_start
        object time_name
        object tterms
        object Ht1, Ht2, Ht3, Ht4
        object d2Ht0, d2Ht1, d2Ht2
        double x0, x1
        object y0, y1
        object d2y0, d2y1
        int _do_update
        object wt
        object wfunc
    cdef unsigned nnz

    def __init__(self, syst, time_name, time_start, params=None, dt=1):

        # The spline interpolation in this routine is done "on the fly".
        # Quite generally, for a cubic spline, the cubic function f(t) is defined
        # on two neighboring intervals [t0, t1] and [t1, t2],
        # t1 - t0 = t2 - t1 = dt. One needs to know
        # f(t) on t0, t1, and t2, and f''(t) on the endpoints t0, t2.
        # As the Schrödinger equation which calls the interpolant propagates only
        # foreward in time, we do a stepwise interpolation and assume, at least
        # for the optimization, that the time argument monontonically increases
        # when the interpolation function is called. The idea is to reuse the
        # evaluated function and derivative values, and don't calculate them twice.
        # Therefore, one always need to keep 5. timesteps in memory, t0, t1 .. t4.
        # Interpolation is always done between t0, t1 and t2.
        # The points t3 and t4 are used to estimate the 2. derivative on t2
        # using a central finite difference scheme with 5 points. The second
        # derivative on point t0 is known from the step before.
        # Only when this routine is initialized, we have to estimate f''(t0).
        # We use a foreward derivative in this case.

        self.wfunc = PerturbationExtractor(syst, time_name, time_start, params)

        # for non-vectorized systems
        _, _, norbs = syst.site_ranges[-1]
        self.size = norbs
        self.nnz = self.wfunc.nnz
        self.time_start = time_start
        self.time_name = time_name
        self.dt = dt

        if dt > 0:

            self.t0 = time_start
            self.t1 = self.t0 + self.dt
            self.t2 = self.t1 + self.dt
            self.t3 = self.t2 + self.dt
            self.t4 = self.t3 + self.dt

            self.Ht0 = self.wfunc.data(self.t0)
            self.Ht1 = self.wfunc.data(self.t1)
            self.Ht2 = self.wfunc.data(self.t2)
            self.Ht3 = self.wfunc.data(self.t3)
            self.Ht4 = self.wfunc.data(self.t4)

            Ht02 = self.wfunc.data(self.t0 + 0.5 * self.dt)
            Ht12 = self.wfunc.data(self.t0 + 1.5 * self.dt)
            Ht22 = self.wfunc.data(self.t0 + 2.5 * self.dt)

            rdt2 = 1 / (self.dt * self.dt)
            rdt2h = 4 * rdt2

            # w''(t0), forward finite difference scheme (4. order)
            # on the half interval dt / 2
            cl0 = 15 / 4 * rdt2h
            cl1 = -77 / 6 * rdt2h
            cl2 = 107 / 6 * rdt2h
            cl3 = -13 * rdt2h
            cl4 = 61 / 12 * rdt2h
            cl5 = -5 / 6 * rdt2h
            self.d2Ht0 = np.zeros((self.nnz,), dtype=np.complex)
            for i in range(self.nnz):
                self.d2Ht0[i] = cl0 * self.Ht0[i] + cl1 * Ht02[i] + cl2 * self.Ht1[i] \
                    + cl3 * Ht12[i] + cl4 * self.Ht2[i] + cl5 * Ht22[i]

            # w''(t2), central finite difference scheme (4. order)
            cc2 = - rdt2 / 12
            cc1 = rdt2 * 4 / 3
            cc0 = - 2.5 * rdt2
            self.d2Ht2 = np.zeros((self.nnz,), dtype=np.complex)
            for i in range(self.nnz):
                self.d2Ht2[i] = cc2 * (self.Ht0[i] + self.Ht4[i]) \
                    + cc1 * (self.Ht1[i] + self.Ht3[i]) \
                    + cc0 * self.Ht2[i]

            # w''(t1), spline formula
            self.d2Ht1 = np.zeros((self.nnz,), dtype=np.complex)
            for i in range(self.nnz):
                self.d2Ht1[i] = 1.5 * rdt2 * (self.Ht0[i] + self.Ht2[i] - 2 * self.Ht1[i]) \
                    - 0.25 * (self.d2Ht0[i] + self.d2Ht2[i])

            self.x0, self.x1 = self.t0, self.t1
            self.y0 = self.Ht0
            self.y1 = self.Ht1
            self.d2y0 = self.d2Ht0
            self.d2y1 = self.d2Ht1
            self._do_update = True

            self.wt = self.wfunc.empty()

    @property
    def size(self):
        '''The size of the W(t) matrix.

        Returns
        -------
        size : int
            The size of the W(t) matrix is ``size`` x ``size``.
        '''
        return self.size

    @property
    def nnz(self):
        '''The total number of time-dependent matrix elements.

        Returns
        -------
        nnz : int
            The total number of time-dependent matrix elements, equal to
            len(W(t).data).
        '''
        return self.nnz

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def evaluate(self, double time, object out=None):
        '''Evaluate the W(t) matrix for the given time t.

        Parameters
        ----------
        time : int or float
            Time argument, must be equal or larger than `time_start`.
        out : `~scipy.sparse.coo_matrix`, optional
            Sparse matrix with to perform the operation in-place.

        Returns
        -------
        out : `~scipy.sparse.coo_matrix` or `None`
            The matrix W(t).
            The size of the W(t) matrix is ``size`` x ``size``, containing
            ``nnz`` complex entries. If an ``out`` argument is given,
            the operation is performed in-place and the routine returns `None`.
        '''

        cdef double c0, c1, c2, rdt2
        cdef double a, b, c, d
        cdef unsigned i

        if out is not None:
            assert out.shape[0] == self.size, _errormsg.format('output',
                                                               out.shape[0],
                                                               self.size)
            assert out.data.shape[0] == self.nnz, _errormsg.format('output.dat',
                                                                   out.data.shape[0],
                                                                   self.nnz)

        if self.dt > 0:

            # update matrices
            if time >= self.t2:
                t0_new, t1_new, t2_new, t3_new, t4_new = update_times(time, self.t0, self.dt)
                rdt2 = 1 / (self.dt * self.dt)
                cc2 = - rdt2 / 12
                cc1 = rdt2 * 4 / 3
                cc0 = - 2.5 * rdt2
                # update matrices H0t, H1t, H2t, try to reuse stuff from last step
                if np.isclose(t0_new, self.t2):
                    for i in range(self.nnz):
                        self.Ht0[i] = self.Ht2[i]
                        self.Ht1[i] = self.Ht3[i]
                        self.Ht2[i] = self.Ht4[i]
                    self.wfunc.data(t3_new, self.Ht3)
                    self.wfunc.data(t4_new, self.Ht4)
                    for i in range(self.nnz):
                        self.d2Ht0[i] = self.d2Ht2[i]
                        # w''(t2), central finite difference scheme
                        self.d2Ht2[i] = cc2 * (self.Ht0[i] + self.Ht4[i]) \
                            + cc1 * (self.Ht1[i] + self.Ht3[i]) \
                            + cc0 * self.Ht2[i]
                else:
                    Htm2 = self.wfunc.data(t0_new - 2 * self.dt)
                    Htm1 = self.wfunc.data(t0_new - self.dt)
                    self.wfunc.data(t0_new, self.Ht0)
                    self.wfunc.data(t1_new, self.Ht1)
                    self.wfunc.data(t2_new, self.Ht2)
                    self.wfunc.data(t3_new, self.Ht3)
                    self.wfunc.data(t4_new, self.Ht4)
                    # w''(t) on t0 and t2, central finite difference scheme
                    for i in range(self.nnz):
                        self.d2Ht0[i] = cc2 * (Htm2[i] + self.Ht2[i]) \
                            + cc1 * (Htm1[i] + self.Ht1[i]) \
                            + cc0 * self.Ht0[i]
                        self.d2Ht2[i] = cc2 * (self.Ht0[i] + self.Ht4[i]) \
                            + cc1 * (self.Ht1[i] + self.Ht3[i]) \
                            + cc0 * self.Ht2[i]

                for i in range(self.nnz):
                    # w''(t) on t1, spline formula
                    self.d2Ht1[i] = 1.5 * rdt2 * (self.Ht0[i] + self.Ht2[i] - 2 * self.Ht1[i])\
                        - 0.25 * (self.d2Ht0[i] + self.d2Ht2[i])
                    self.y0[i] = self.Ht0[i]
                    self.y1[i] = self.Ht1[i]
                    self.d2y0[i] = self.d2Ht0[i]
                    self.d2y1[i] = self.d2Ht1[i]
                self.t0, self.t1, self.t2, self.t3, self.t4 = t0_new, t1_new, t2_new, t3_new, t4_new
                self.x0, self.x1 = self.t0, self.t1
                self._do_update = True

            # to construct the spline, we need f and f''
            # on the two intervals [t0, t1] and [t1, t2].
            # the spline is then evaluated first on the left, then on the right
            # interval. again f and f'' are shifted from left to right interval.
            if self._do_update and time > self.t1:
                for i in range(self.nnz):
                    self.y0[i] = self.Ht1[i]
                    self.y1[i] = self.Ht2[i]
                    self.d2y0[i] = self.d2Ht1[i]
                    self.d2y1[i] = self.d2Ht2[i]
                self.x0, self.x1 = self.t1, self.t2
                self._do_update = False

            a, b, c, d = _get_polynome_coeffs(time, self.x0, self.x1, self.dt)

            if out is None:
                for i in range(self.nnz):
                    self.wt.data[i] = a * self.y0[i] + b * self.y1[i] \
                        + c * self.d2y0[i] + d * self.d2y1[i]
                return self.wt
            else:
                for i in range(self.nnz):
                    out.data[i] = a * self.y0[i] + b * self.y1[i] \
                        + c * self.d2y0[i] + d * self.d2y1[i]

        else:
            if out is None:
                do_return = True
                out = self.wfunc.evaluate(time)
            else:
                do_return = False
                self.wfunc.evaluate(time, out)
            return out if do_return else None

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def apply(self, double time, const complex[:] ket, object out=None):
        '''Evaluate the matrix-vector product W(t) @ ket for the given time.

        Parameters
        ----------
        time : int or float
            Time argument, must be equal or larger than `time_start`.
        out : `numpy.ndarray`, optional
            Sparse matrix with to perform the operation in-place.

        Returns
        -------
        out : `numpy.ndarray` or `None`
            The product W(t) @ ket. If an ``out`` argument is given,
            the operation is performed in-place and the routine returns `None`.
        '''

        do_return = out is None

        if self.dt > 0:

            assert ket.shape[0] == self.size, _errormsg.format('Ket',
                                                               ket.shape[0],
                                                               self.size)
            if out is None:
                out = np.zeros((self.size,), dtype=np.complex)
            else:
                assert out.shape[0] == self.size, _errormsg.format('output',
                                                                   out.shape[0],
                                                                   self.size)

            self.wt = self.evaluate(time)

            coo_matvec(
                self.wt.nnz, self.wt.row, self.wt.col, self.wt.data,
                ket, out)

        else:
            if out is None:
                do_return = True
                out = self.wfunc.apply(time, ket)
            else:
                do_return = False
                self.wfunc.apply(time, ket, out)

        return out if do_return else None

# Example Kernel implementation in C

cdef class _CSRMatrix:
    cdef complex[:] data
    cdef int nrows, ncols
    cdef int[:] indices, indptr

    def __cinit__(self, mat):
        mat = mat.tocsr()
        self.nrows, self.ncols = mat.shape
        self.data = mat.data
        self.indices, self.indptr = mat.indices, mat.indptr


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef _csr_spmv(_CSRMatrix A, const complex *X, complex *Y):
    """Naive implementation of CSR-format matrix-vector multiplication.

    This is taken almost verbatim from scipy.sparse.
    """
    cdef complex sum
    cdef unsigned i, jj

    for i in range(A.nrows):
       sum = Y[i]
       for jj in range(A.indptr[i], A.indptr[i+1]):
           sum += A.data[jj] * X[A.indices[jj]]
       Y[i] = sum


cdef class Simple(CKernel):
    """A C-implemented kernel using naive CSR matvec."""

    cdef:
        object W
        object params
        int center_size
        complex[:] _psi_st, _temp
        complex *temp,
        complex *psi_st
        _CSRMatrix H0

    def __init__(self, H0, W, params, complex[:] psi_st=None):
        pass

    def set_params(self, params):
        self.params = params

    def __cinit__(self, H0, W, params, complex[:] psi_st=None):
        self.c_self = <void*> self
        self.params = params
        self.c_rhs = self._rhs
        self.W = W
        self.size = H0.shape[0]
        self._psi_st = psi_st
        if psi_st is None:
            self.psi_st = NULL
        else:
            self.psi_st = &psi_st[0]
        if W is not None:
            self.center_size = self.W.size
        else:
            self.center_size = self.size
        self.H0 = _CSRMatrix(H0)
        self._temp = np.empty((self.center_size,), dtype=complex)
        self.temp = &self._temp[0]

    # RHS implemented as a static method to conform to the solver interface.
    @staticmethod
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef void _rhs(const void *_self, const complex *psi, complex *dpsidt,
                   double time) except *:
        cdef Simple self = <Simple>_self
        cdef unsigned i
        for i in range(self.size):
            dpsidt[i] = 0
        ### H0 @ psi -> dpsidt
        _csr_spmv(self.H0, psi, dpsidt)
        ### W(t) @ psi + dpsidt -> dpsidt
        if self.psi_st == NULL:
            try:
                self.W.apply(time, <complex[:self.center_size]>psi,
                             out=<complex[:self.center_size]>dpsidt)
            except Exception as exept:
                if self.W is None:
                    pass
                else:
                    raise exept
        else:
            # we need to act on `temp = psi + psi_st`
            # use pre-allocated storage
            for i in range(self.center_size):
                self.temp[i] = psi[i] + self.psi_st[i]
            try:
                self.W.apply(time, <complex[:self.center_size]>self.temp,
                             out=<complex[:self.center_size]>dpsidt)
            except Exception as exept:
                if self.W is None:
                    pass
                else:
                    raise exept
        ### dpsidt = -1j * dpsidt
        for i in range(self.size):
            dpsidt[i] = -1j * dpsidt[i]


try:
    from ._sparse_blas_kernel import SparseBlas
    __all__.append('SparseBlas')
except ImportError as e:
    pass

default = Scipy



import time as timer
import numpy as np
from matplotlib import pyplot as plt

import kwant
import tkwant


def make_system(a=1, gamma=1.0, radius=10):
    """Make a tight binding system on a single square lattice"""
    # `a` is the lattice constant and `gamma` the hopping integral
    # both set by default to 1 for simplicity.

    lat = kwant.lattice.square(a, norbs=1)

    syst = kwant.Builder()

    # Define the quantum dot
    def circle(pos):
        (x, y) = pos
        return x ** 2 + y ** 2 < radius ** 2

    syst[lat.shape(circle, (0, 0))] = 4 * gamma
    syst[lat.neighbors()] = -gamma

    lead = kwant.Builder(kwant.TranslationalSymmetry((-1, 0)))
    lead[(lat(0, j) for j in range(-radius // 2 + 2, radius // 2 - 1))] = 4 * gamma
    lead[lat.neighbors()] = -gamma

    syst.attach_lead(lead)
    syst.attach_lead(lead.reversed())

    return syst


# for adding a time-dependent voltage on top of the leads
def faraday_flux(time):
    return 0.05 * (time - 10 * np.sin(0.1 * time))


def evolve(times, psi, operator):
    ops = []
    for time in times:
        psi.evolve(time)
        expectation = psi.evaluate(operator)
        ops.append(expectation)
    return ops


def main():
    syst = make_system()

    # add a time-dependent voltage to lead 0 -- this is implemented
    # by adding sites to the system at the interface with the lead and
    # multiplying the hoppings to these sites by exp(-1j * faraday_flux(time))
    extra_sites = tkwant.leads.add_voltage(syst, 0, faraday_flux)
    lead_syst_hoppings = [(s, site) for site in extra_sites
                          for s in syst.neighbors(site)
                          if s not in extra_sites]

    kwant.plot(syst, site_color='k', lead_color='grey', num_lead_cells=4,
               hop_lw=lambda a, b: 0.3 if (a, b) in lead_syst_hoppings else 0.1,
               hop_color=lambda a, b: 'r' if (a, b) in lead_syst_hoppings else 'k')

    syst = syst.finalized()

    # create an observable for calculating the current flowing from the left lead
    current_operator = kwant.operator.Current(syst, where=lead_syst_hoppings,
                                              sum=True)

    energy = 1.
    tmax = 600

    # create initial scattering state
    scattering_states = kwant.wave_function(syst, energy=energy, params={'time': 0})
    psi_st = scattering_states(0)[0]

    # boundary conditions typically used by `tkwant.solve`
    simple_boundaries = [tkwant.leads.SimpleBoundary(tmax=tmax)
                         for l in syst.leads]

    # boundary conditions with an absorbing potential that increases
    # according to x**n where `x` is the distance into the lead
    absorbing_boundaries = [tkwant.leads.MonomialAbsorbingBoundary(num_cells=100,
                                                                   strength=10,
                                                                   degree=6)
                            for l in syst.leads]

    # create time-dependent wavefunctions that starts in a scattering state
    # originating from the left lead, using two different types of boundary
    # conditions
    psi_simple = tkwant.onebody.WaveFunction.from_kwant(syst=syst, psi_init=psi_st,
                                                        boundaries=simple_boundaries,
                                                        energy=energy)

    psi_alt = tkwant.onebody.WaveFunction.from_kwant(syst=syst, psi_init=psi_st,
                                                     boundaries=absorbing_boundaries,
                                                     energy=energy)

    # evolve forward in time, calculating the current
    times = np.arange(0, tmax)

    # Simple boundary conditions
    start = timer.process_time()
    current_simple = evolve(times, psi_simple, current_operator)
    stop = timer.process_time()
    print('Simple Boundary conditions elapsed time: {:.2f}s'.format(stop - start))

    # Absorbing boundary conditions
    start = timer.process_time()
    current_alt = evolve(times, psi_alt, current_operator)
    stop = timer.process_time()
    print('Absorbing Boundary conditions elapsed time: {:.2f}s'.format(stop - start))

    plt.plot(times, current_simple, lw=2, label='SimpleBoundary')
    plt.plot(times, current_alt, '--', lw=2, label='MonomialAbsorbingBoundary')
    plt.xlabel(r'time $t$')
    plt.ylabel(r'current $I(t)$')
    plt.legend()
    plt.show()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
#
# cython: embedsignature=True
#
# Copyright 2016, 2017 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.


cdef class Kernel:
    cdef public int size, nevals


ctypedef void (*rhs_t)(const void *self, const complex *psi_bar, complex *dpsidt,
                       double time) except *


cdef class CKernel(Kernel):
    cdef void *c_self
    cdef rhs_t c_rhs



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# tkwant documentation build configuration file, created by
# sphinx-quickstart on Sun May 22 14:08:28 2016.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# -- Path setup --------------------------------------------------------------

# The path is set in the Makefile, so that it also affects the Jupyter kernel
# executed through the jupyter-sphinx extension.

import sys
import os

# For kwantdoc extenstion
sys.path.insert(0, os.path.abspath('../sphinxext'))

import tkwant

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
    'jupyter_sphinx.execute',
    'kwantdoc',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['../templates']

# The suffix(es) of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'tkwant'
author = 'J. Weston (TUD, CEA), T. Kloss (HLRS, CEA), C. W. Groth (CEA), X. Waintal (CEA), and others'
copyright = '2016-2019, ' + author

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version, including alpha/beta/rc tags.
release = tkwant.__version__

# The short X.Y version.
version = release[:len(release) - len(release.lstrip('012345679.'))].rstrip('.')

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build']

# The reST default role (used for this markup: `text`) to use for all documents.
default_role = "autolink"

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
#keep_warnings = False

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# http://stackoverflow.com/questions/9728292/creating-latex-math-macros-within-sphinx
pngmath_latex_preamble = r"""\newcommand{\bra}[1]{\left\langle#1\right|}
\newcommand{\ket}[1]{\left|#1\right>}
\newcommand{\braket}[2]{\left\langle#1|#2\right\rangle}
\newcommand{\ri}{\text{i}}
\newcommand{\rd}{\text{d}}
"""

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
html_theme = 'kwantdoctheme'
html_theme_path = ['..']
html_theme_options = {'collapsiblesidebar': True}
html_style = 'kwantdoctheme.css'


# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.
# "<project> v<release> documentation" by default.
#html_title = 'tkwant v0'

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#html_logo = None

# The name of an image file (relative to this directory) to use as a favicon of
# the docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
#html_extra_path = []

# If not None, a 'Last updated on:' timestamp is inserted at every page
# bottom, using the given strftime format.
# The empty string is equivalent to '%b %d, %Y'.
#html_last_updated_fmt = None

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Language to be used for generating the HTML full-text search index.
# Sphinx supports the following languages:
#   'da', 'de', 'en', 'es', 'fi', 'fr', 'h', 'it', 'ja'
#   'nl', 'no', 'pt', 'ro', 'r', 'sv', 'tr', 'zh'
#html_search_language = 'en'

# A dictionary with options for the search language support, empty by default.
# 'ja' uses this config value.
# 'zh' user can custom change `jieba` dictionary path.
#html_search_options = {'type': 'default'}

# The name of a javascript file (relative to the configuration directory) that
# implements a search results scorer. If empty, the default will be used.
#html_search_scorer = 'scorer.js'

# Output file base name for HTML help builder.
htmlhelp_basename = 'tkwantdoc'

# -- Options for LaTeX output ---------------------------------------------
# http://thread.gmane.org/gmane.comp.python.sphinx.devel/4220/focus=4238
latex_elements = {'papersize': 'a4paper',
                  'release': '',
                  'releasename': '',
                  'preamble':
r"""\makeatletter
  \fancypagestyle{normal}{
    \fancyhf{}
    \fancyfoot[LE,RO]{{\py@HeaderFamily\thepage}}
    \fancyfoot[LO]{{\py@HeaderFamily\nouppercase{\rightmark}}}
    \fancyfoot[RE]{{\py@HeaderFamily\nouppercase{\leftmark}}}
    \fancyhead[LE,RO]{{\py@HeaderFamily \@title}}
    \renewcommand{\headrulewidth}{0.4pt}
    \renewcommand{\footrulewidth}{0.4pt}
  }
\makeatother
""" + pngmath_latex_preamble}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'tkwant.tex', 'tkwant {0} documentation'.format(release),
     author, 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'tkwant', 'tkwant Documentation',
     [author], 1)
]

# If true, show URL addresses after external links.
#man_show_urls = False


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'tkwant', 'tkwant Documentation',
     author, 'tkwant', 'One line description of project.',
     'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
#texinfo_no_detailmenu = False

# -- Options for autodoc -------------------------------------------------------
# Generate stub pages for autosummary directives.
autosummary_generate = True

autoclass_content = "both"
autodoc_default_flags = ['show-inheritance']

# -- Teach Sphinx to document bound methods like functions ---------------------
import types
from sphinx.ext import autodoc

class BoundMethodDocumenter(autodoc.FunctionDocumenter):
    objtype = "boundmethod"
    directivetype = 'function'

    @classmethod
    def can_document_member(cls, member, membername, isattr, parent):
        # Return True iff `member` is a bound method.  Taken from
        # <http://stackoverflow.com/a/1260881>.
        return (isinstance(member, types.MethodType) and
                member.__self__ is not None and
                not issubclass(member.__self__.__class__, type) and
                member.__self__.__class__ is not type)

    def format_args(self):
        args = super(BoundMethodDocumenter, self).format_args()
        left, sep, right = args.partition('self, ')
        if left.endswith('('):
            args = left + right
        return args

def setup(app):
    app.add_autodocumenter(BoundMethodDocumenter)

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'python': ('https://docs.python.org/3.4/', None),
    'numpy': ('https://docs.scipy.org/doc/numpy/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference/', None),
    'mpi4py': ('https://mpi4py.readthedocs.io/en/stable/', None),
    'kwant': ('https://kwant-project.org/doc/1/', None),
    'kwantspectrum': ('https://kwant-project.org/extensions/kwantspectrum/', None)
}



# -*- coding: utf-8 -*-
# Copyright 2016-2020 tkwant authors.
#
# This file is part of tkwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of tkwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

"""Test module for routines using MPI"""

import pytest
from pytest import raises
import cmath
import numpy as np
import kwant
from numpy.testing import assert_array_almost_equal
from scipy.special import erf
import warnings


from . import mpi_testdriver
from .. import onebody, manybody, system, leads, integration, mpi
import kwantspectrum

pytestmark = pytest.mark.mpitest  # module level marker for mpi tests
num_processes = 4


def make_simple_lead(W=3):
    lat = kwant.lattice.square(a=1, norbs=1)
    sym = kwant.TranslationalSymmetry((-1, 0))
    H = kwant.Builder(sym)
    H[(lat(0, y) for y in range(W))] = 0
    H[lat.neighbors()] = -1
    return H


def onsite_potential(site, time, v):
    return 4 * (1 + v * cmath.exp(- 1j * time))


def gaussian(time):
    t0 = 0
    A = 0.00157
    sigma = 24
    return A * (1 + erf((time - t0) / sigma))


# time dependent coupling with gaussian pulse
def coupling_nn(site1, site2, time):
    return - cmath.exp(- 1j * gaussian(time))


def make_system(a=1, gamma=1.0, W=1, L=30):

    lat = kwant.lattice.square(a=a, norbs=1)
    syst = kwant.Builder()

    # Define the scattering region.
    syst[(lat(x, y) for x in range(L) for y in range(W))] = onsite_potential
    syst[lat.neighbors()] = -gamma
    # time dependent coupling between two sites 0 and 1
    for y in range(W):
        syst[lat(0, y), lat(1, y)] = coupling_nn

    # Define and attach the leads.
    # Construct the left lead.
    lead = kwant.Builder(kwant.TranslationalSymmetry((-a, 0)))
    lead[(lat(0, j) for j in range(W))] = 4 * gamma
    lead[lat.neighbors()] = -gamma

    # Attach the left lead and its reversed copy.
    syst.attach_lead(lead)
    syst.attach_lead(lead.reversed())

    return syst


class ManybodyIntegrand:

    def __init__(self, syst, interval, operator, time=0, params=None, element=None):
        """Diagnostic routine to evaluate the integrand in the manybody integral"""

        tparams = system.add_time_to_params(params, time_name='time',
                                            time=0, check_numeric_type=True)

        spectra = kwantspectrum.spectra(syst.leads, params=tparams)
        self.boundaries = leads.automatic_boundary(spectra, tmax=time)

        self.syst = syst
        self.band = interval.band
        self.lead = interval.lead
        self.time = time
        self.operator = operator
        self.params = params
        self.spectrum = spectra[self.lead]
        self.element = element

    def _f(self, k):
        """integrand (momentum integration) for the manybody expectation value
        at zero temperature"""
        energy = self.spectrum(k, band=self.band)
        velocity = self.spectrum(k, band=self.band, derivative_order=1)
        mode = self.spectrum.momentum_to_scattering_mode(k, self.band)

        psi = onebody.ScatteringStates(self.syst, energy=energy, lead=self.lead,
                                       params=self.params,
                                       boundaries=self.boundaries)[mode]

        psi.evolve(self.time)

        result = velocity * psi.evaluate(self.operator) / (2 * np.pi)
        if result.shape == (1,):  # if result is an array with only one element
            result = result[0]
        if self.element is None:
            return result
        return result[self.element]

    def func(self, k):
        return np.array([self._f(xi) for xi in k])


def quadpack_error_per_interval(func, order, ai, bi):

    def calc_ik(x, w, kronrod):
        integrand = np.abs(func(x) - kronrod / (bi - ai))
        return np.sum(w * integrand, axis=1)[1]

    x, w = integration.calc_abscissas_and_weights(ai, bi, order, quadrature='kronrod')

    gauss, kronrod = np.sum(w * func(x), axis=1)
    ik = calc_ik(x, w, kronrod)
    tmp = 200 * np.abs(gauss - kronrod) / ik
    return ik * np.minimum(1, tmp * np.sqrt(tmp))


def quadpack_qag(func, a, b, atol=1.49e-8, rtol=1.49e-8, limit=50, order=5):
    # import sys

    def estimate_interval(ai, bi):
        x, w = integration.calc_abscissas_and_weights(ai, bi, order, quadrature='kronrod')
        return np.sum(w * func(x), axis=1)[1]

    def estimate_full_integral(intervals):
        return [estimate_interval(*interval) for interval in intervals]

    def estimate_errors(intervals):
        return np.array([quadpack_error_per_interval(func, order, *interval)
                         for interval in intervals])

    intervals = [(a, b)]

    i = 1
    while True:  # loop until converged

        # print('cycle {}'.format(i))

        results = estimate_full_integral(intervals)

        result = np.sum(results)

        tol = max(atol, rtol * result)

        errors = estimate_errors(intervals)

        max_error_per_interval = [np.max(error - tol) for error in errors]

        max_index = np.argsort(max_error_per_interval)[::-1]

        intervals = [intervals[i] for i in max_index]
        results = [results[i] for i in max_index]
        errors = errors[max_index]

        # for inte, res, err in zip(intervals, results, errors):
        #     print(inte, float(res), float(err))
        # print('errorsum, errorbnd {} {}'.format(float(np.sum(errors)), tol))
        # sys.stdout.flush()

        if np.where(errors < tol, True, False).all():
            break

        i += 1
        if i > limit:
            warnings.warn('maxdepth for interval refinement exceeded')
            break

        ai, bi = intervals.pop(0)
        mi = (ai + bi) / 2

        intervals.append((ai, mi))
        intervals.append((mi, bi))

    return result, errors, max_error_per_interval, intervals


@mpi_testdriver.wrap(num_processes)
def test_mpi_communicator_init():

    assert mpi._COMM is None
    mpi.communicator_init()
    assert mpi._COMM is not None
    from mpi4py import MPI
    assert MPI.Comm.Compare(mpi._COMM, MPI.COMM_WORLD) == MPI.CONGRUENT

    # check reinitialization
    # TODO: it would be nice if we could check the warning message
    mpi.communicator_init()


@mpi_testdriver.wrap(num_processes)
def test_mpi_communicator_init_with_argument():

    assert mpi._COMM is None

    # split COMM_WORLD in two communicators of half size with colour 0 and 1
    from mpi4py import MPI
    colour = MPI.COMM_WORLD.rank // 2
    key = 0
    new_comm = MPI.COMM_WORLD.Split(colour, key)
    assert new_comm.size == MPI.COMM_WORLD.size / 2

    mpi.communicator_init(new_comm)

    assert mpi._COMM is not None
    assert MPI.Comm.Compare(mpi._COMM, new_comm) == MPI.CONGRUENT
    assert MPI.Comm.Compare(mpi._COMM, MPI.COMM_WORLD) == MPI.UNEQUAL


@mpi_testdriver.wrap(num_processes)
def test_mpi_get_communicator():

    comm = mpi.get_communicator()
    assert comm is not None
    from mpi4py import MPI
    assert MPI.Comm.Compare(comm, MPI.COMM_WORLD) == MPI.CONGRUENT

    # split COMM_WORLD in two communicators of half size with colour 0 and 1
    colour = MPI.COMM_WORLD.rank // 2
    key = 0
    new_comm = MPI.COMM_WORLD.Split(colour, key)
    assert new_comm.size == MPI.COMM_WORLD.size / 2

    # check that passing a communicator only return the same communicator again
    comm_1 = mpi.get_communicator(new_comm)
    assert MPI.Comm.Compare(new_comm, comm_1) == MPI.IDENT

    # check that the tkwant global communicator is still the same
    comm_2 = mpi.get_communicator()

    assert MPI.Comm.Compare(comm_2, comm) == MPI.IDENT
    assert MPI.Comm.Compare(comm_2, MPI.COMM_WORLD) == MPI.CONGRUENT
    assert MPI.Comm.Compare(comm_2, comm_1) == MPI.UNEQUAL


@mpi_testdriver.wrap(num_processes)
def test_mpi_get_communicator_does_not_set_state():

    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    comm_1 = mpi.get_communicator(comm)
    assert MPI.Comm.Compare(comm, comm_1) == MPI.IDENT
    assert mpi._COMM is None


@mpi_testdriver.wrap(num_processes)
def test_mpi_round_robin():

    comm = mpi.get_communicator()
    rank = comm.rank
    size = comm.size

    assert size >= 4

    # test condition: rank == key + m * size

    for key in range(size):  # m = 0
        if rank == key:
            assert mpi.round_robin(comm, key)
        else:
            assert not mpi.round_robin(comm, key)

    for key in range(size, 2 * size):  # m = -1
        if rank == key - size:
            assert mpi.round_robin(comm, key)
        else:
            assert not mpi.round_robin(comm, key)

    for key in range(- size, 0):  # m = 1
        if rank == key + size:
            assert mpi.round_robin(comm, key)
        else:
            assert not mpi.round_robin(comm, key)


@mpi_testdriver.wrap(num_processes)
def test_mpi_distribute_dict():

    comm = mpi.get_communicator()
    rank = comm.rank
    size = comm.size

    assert size >= 4

    non_distributed_dict = {'a': 5, 'b': 8}

    def map_keys_to_ranks(commun, key):
        # map a -> rank 1 and b to rank 0
        if (key == 'a' and commun.rank == 1) or (key == 'b' and commun.rank == 0):
            return True
        return False

    distributed_dict = mpi.distribute_dict(non_distributed_dict,
                                           map_keys_to_ranks, comm)

    assert isinstance(distributed_dict, dict)

    if rank == 0:
        assert [('b', 8)] == list(distributed_dict.items())
    if rank == 1:
        assert [('a', 5)] == list(distributed_dict.items())
    if rank > 1:
        assert len(distributed_dict) == 0


@mpi_testdriver.wrap(num_processes)
def test_distributed_dict():

    comm = mpi.get_communicator()
    rank = comm.rank
    size = comm.size

    assert size >= 4

    if rank == 3:
        data = {'a': 7, 'b': 3.14}
    elif rank == 2:
        data = {'c': 7}
    else:
        data = {}

    non_present_key = 'd'
    non_present_rank = size

    distributed_dict = mpi.DistributedDict(data)

    # add data
    distributed_dict.add(key=5, value=1)  # add data somewhere
    assert set(distributed_dict.keys()) == set(['a', 'b', 'c', 5])
    raises(ValueError, distributed_dict.add, key=5, value=1)  # key exists already
    distributed_dict.add(key=24, value=1, rank=3)  # add data on a specific rank
    assert set(distributed_dict.keys(rank=3)) == set(['a', 'b', 24])
    raises(ValueError, distributed_dict.add, key=24, value=1, rank=3)
    raises(ValueError, distributed_dict.add, key=25, value=1, rank=non_present_rank)

    # delete data
    distributed_dict.delete(key=5)
    assert set(distributed_dict.keys()) == set(['a', 'b', 'c', 24])
    raises(ValueError, distributed_dict.delete, key=non_present_key)

    # get data
    element = distributed_dict.data(key='a')
    assert element == 7
    raises(ValueError, distributed_dict.data, key=non_present_key)

    # get local data
    local_data = distributed_dict.local_data()
    element_inside = [value == local_data[key] for key, value in data.items()]
    assert np.array(element_inside).all()

    # get local keys
    keys = distributed_dict.local_keys()
    if rank == 3:
        assert set(keys) == set(['a', 'b', 24])
    if rank == 2:
        assert set(keys) == set(['c'])

    # get keys
    keys = distributed_dict.keys()  # all keys
    assert isinstance(keys, list)
    assert set(keys) == set(['a', 'b', 'c', 24])
    keys = distributed_dict.keys(rank=3)  # keys on specific ranks
    assert isinstance(keys, list)
    assert set(keys) == set(['a', 'b', 24])
    keys = distributed_dict.keys(rank=2)
    assert isinstance(keys, list)
    assert set(keys) == set(['c'])
    raises(AssertionError, distributed_dict.keys, rank=non_present_rank)

    # get size
    dict_size = distributed_dict.size()  # size on all ranks
    assert isinstance(dict_size, np.ndarray)
    assert (dict_size == [0, 0, 1, 3]).all()
    assert distributed_dict.size(rank=0) == 0  # size on a specific rank
    assert distributed_dict.size(rank=1) == 0
    assert distributed_dict.size(rank=2) == 1
    assert distributed_dict.size(rank=3) == 3
    raises(AssertionError, distributed_dict.size, rank=non_present_rank)

    # move data
    distributed_dict.move_data(key='b', to_rank=2)  # move from rank 3 -> rank 2
    assert set(distributed_dict.keys(rank=3)) == set(['a', 24])
    assert set(distributed_dict.keys(rank=2)) == set(['b', 'c'])
    # distributed_dict.move_data(key='b', to_rank=2)  # move from rank 2 -> rank 2
    # TODO: does not work

    raises(ValueError, distributed_dict.move_data, key=non_present_key, to_rank=2)
    # raises(AssertionError, distributed_dict.move_data, key='a', to_rank=non_present_rank)
    # TODO: get error from mpi4py

    # get rank number
    assert distributed_dict.rank_from_key(key='a') == 3
    assert distributed_dict.rank_from_key(key=24) == 3
    assert distributed_dict.rank_from_key(key='b') == 2
    assert distributed_dict.rank_from_key(key='c') == 2
    raises(ValueError, distributed_dict.rank_from_key, key=non_present_key)

    # check if key is present
    for key in ['a', 'b', 'c', 24]:
        assert distributed_dict.key_is_present(key=key)
    assert not distributed_dict.key_is_present(key='bla')

    # check if keys are unique
    assert distributed_dict.keys_are_unique()
    if rank == 1:
        distributed_dict._data = {'c': 7}  # add data already present on rank 2
    assert not distributed_dict.keys_are_unique()

    # wrong input with dublicated data
    if rank == 3:
        dublicated_data = {'a': 7, 'b': 3.14}
    elif rank == 2:
        dublicated_data = {'b': 3.14}
    else:
        dublicated_data = {}

    raises(ValueError, mpi.DistributedDict, data=dublicated_data)


@mpi_testdriver.wrap(num_processes)
def test_manybody_wavefunction():

    def test_wavefunction(number_energy_points):

        time = 5
        energies = np.linspace(2.1, 3, number_energy_points)

        params = {'v': 0.0001}

        # ------------- test evolve and evaluate --------------------------

        # compute in a non-mpi fashion the manybody density by hand
        # we take a weight for each onebody contribution to be 1 for simplicity
        psis_ref = [onebody.ScatteringStates(syst, energy=energy, lead=0, tmax=50,
                                             params=params)[0]
                    for energy in energies]

        for psi in psis_ref:
            psi.evolve(time)

        density_ref = sum([psi.evaluate(density_operator)
                           for psi in psis_ref])

        # compute the same density with a MPI distributed wavefunction
        class MyTask:
            weight = np.array([1])

        psi_init = {}
        for i, energy in enumerate(energies):
            if rank == i % size:
                psi_init[i] = onebody.ScatteringStates(syst, energy=energy,
                                                       lead=0, tmax=50, params=params)[0]
        tasks = {i: MyTask() for i in range(len(energies))}

        # compare the densities for consistency
        wavefunction = manybody.WaveFunction(psi_init, tasks)
        wavefunction.evolve(time)
        density = wavefunction.evaluate(density_operator)
        if rank == 0:  # evaluate returns only on rank 0
            assert_array_almost_equal(density, density_ref)

        # --------------------- test _calc_expectation_value --------------
        # over all keys
        density = wavefunction._calc_expectation_value(density_operator)
        if rank == 0:  # _calc_expectation_value returns only on rank 0
            assert_array_almost_equal(density, density_ref)
        # partial evaluate over a subset of keys
        keys = np.array(list(tasks.keys()))[1::2]
        if len(keys) > 0:
            density = wavefunction._calc_expectation_value(density_operator, keys=keys)
            density_ref = sum([psis_ref[key].evaluate(density_operator) for key in keys])
            if rank == 0:  # _calc_expectation_value returns only on rank 0
                assert_array_almost_equal(density, density_ref)

        # --------------------- test get onebody state-----------------------
        for key in range(len(energies)):
            psi = wavefunction.get_onebody_state(key)
            psi_ref = psis_ref[key]
            assert np.allclose(psi.psi(), psi_ref.psi())

        # --------------------- test to add onebody state---------------------
        new_energy = 2.37423  # some arbitray number where one open mode exists
        new_state = onebody.ScatteringStates(syst, energy=new_energy, lead=0, tmax=50, params=params)[0]

        psis_ref.append(new_state)
        time += 5
        for psi in psis_ref:
            psi.evolve(time)
        density_ref = sum([psi.evaluate(density_operator) for psi in psis_ref])

        new_key = wavefunction.add_onebody_state(state=new_state, task=MyTask)
        new_state = None  # delete in order to make sure that we have not created any hidden reference
        wavefunction.evolve(time)
        density = wavefunction.evaluate(density_operator)
        if rank == 0:  # evaluate returns only on rank 0
            assert_array_almost_equal(density, density_ref)

        assert isinstance(new_key, int)
        assert new_key == number_energy_points

        new_state = onebody.ScatteringStates(syst, energy=new_energy, lead=0, tmax=50, params=params)[0]
        new_key = wavefunction.add_onebody_state(state=new_state, task=MyTask, key='bla')
        assert new_key == 'bla'

        raises(ValueError, wavefunction.add_onebody_state,
               state=new_state, task=MyTask, key='bla')  # key exists already

        # ----------------- test to delete onebody state--------------------
        wavefunction.delete_onebody_state(key='bla')
        raises(ValueError, wavefunction.delete_onebody_state, key='bla')  # already deleted

        # -------------------- get all keys----------------------------
        keys = wavefunction.get_keys()
        assert isinstance(keys, list)
        assert set(keys) == set(range(number_energy_points + 1))  # we have added one state above

        # --------------------  get a free key--------------------
        assert wavefunction.get_free_key() == number_energy_points + 1
        wavefunction.add_onebody_state(state=new_state, task=MyTask, key='bla')
        assert wavefunction.get_free_key() == number_energy_points + 1
        wavefunction.add_onebody_state(state=new_state, task=MyTask, key=number_energy_points + 5)
        assert wavefunction.get_free_key() == number_energy_points + 6

        # --------------------  test weight shape--------------------
        assert wavefunction.weight_shape() == (1, )

        # --------------------  test consistency--------------------
        assert wavefunction._check_consistency()
        del wavefunction.tasks[0]
        assert not wavefunction._check_consistency()

    comm = mpi.get_communicator()
    rank = comm.rank
    size = comm.size
    assert size > 1

    syst = make_system().finalized()
    density_operator = kwant.operator.Density(syst)

    # test for consistency in the case that we have less, equal, or more
    # modes that MPI ranks
    test_wavefunction(number_energy_points=size // 2)
    test_wavefunction(number_energy_points=size)
    test_wavefunction(number_energy_points=size + size // 2)


@mpi_testdriver.wrap(num_processes)
def test_manybody_state():

    comm = mpi.get_communicator()
    rank = comm.rank
    assert comm.size > 1

    syst = make_system().finalized()
    density_operator = kwant.operator.Density(syst)

    tmax = 50
    time = 5
    params = {'v': 0.0001}

    occupation = manybody.lead_occupation(chemical_potential=3)

    # manybody solver
    state = manybody.State(syst, tmax, occupation, params=params, refine=False)

    # check that result is similar to reference result
    state.evolve(time)

    # -- test refine_intervals
    error_before = state.estimate_error()
    state.refine_intervals(atol=1E-3, rtol=1E-3)
    error_after = state.estimate_error()
    assert error_after < error_before

    density = state.evaluate(density_operator)

    # get the tasks and recalculate non-mpi the same
    tasks = [task for _, task in state.manybody_wavefunction.tasks.items()]
    psis_ref = [onebody.ScatteringStates(syst, energy=task.energy,
                                         lead=task.lead, tmax=tmax, params=params)[task.mode]
                for task in tasks]

    for psi in psis_ref:
        psi.evolve(time)

    density_ref = sum([task.weight[-1] * psi.evaluate(density_operator)
                       for task, psi in zip(tasks, psis_ref)])

    if rank == 0:  # evaluate returns only on rank 0
        assert_array_almost_equal(density, density_ref)


@mpi_testdriver.wrap(num_processes)
def test_quadpack_error_estimate():

    comm = mpi.get_communicator()
    assert comm.size > 1

    syst = make_system().finalized()

    element = 10  # the site where do we want to compare the error

    def site_pos(site):
        return site.pos[0] == element

    density_operator = kwant.operator.Density(syst, where=site_pos)  # operator with only one element
    density_operator_array = kwant.operator.Density(syst)  # operator with array output

    time = 5
    tmax = time
    params = {'v': 0.0001}
    occupation = manybody.lead_occupation(chemical_potential=3)

    state = manybody.State(syst, tmax, occupation, params=params, refine=False)

    # estimate the error at a given time in the first interval for a given element
    intervals = state.get_intervals()
    interval = intervals[0]

    state.evolve(time)

    # compare if array valued operator and single-valued operator give the same result
    error = state._error_estimate_quadpack(interval=interval, error_op=density_operator)

    error_array = state._error_estimate_quadpack(interval=interval,
                                                 error_op=density_operator_array)

    assert np.allclose(error, error_array[element])

    # compare the error to the quadpack reference implementation
    integrand = ManybodyIntegrand(syst, interval=interval, time=time, params=params,
                                  operator=density_operator)

    reference_error = quadpack_error_per_interval(integrand.func, interval.order,
                                                  interval.kmin, interval.kmax)

    assert np.allclose(error, reference_error)


@mpi_testdriver.wrap(num_processes)
def test_quadpack_adaptive_refinement_of_manybody_state():

    # we test the adaptive refinement against a simple reference implementation
    # (scipy quadpack does not implement the QAG, but only the QAGS algorithm)
    # moreover, we test our routine for two different operators.
    # first, a density operator where we measure the density at only one site,
    # such that the density has a scalar value.
    # second a density operator, where the density is site dependent. to compare
    # both cases, we set the array density to the scalar density value everywhere
    # the result should both match the scalar, as well as the reference result

    comm = mpi.get_communicator()
    assert comm.size > 1

    syst = make_system().finalized()

    element = 10  # the site where do we want to compare the error

    def site_pos(site):
        return site.pos[0] == element

    density_operator = kwant.operator.Density(syst, where=site_pos)  # operator with only one element
    density_operator_array = kwant.operator.Density(syst)  # operator with array output

    def make_operator(operator):
        """make an operator where all elements are equal **element**"""
        def homogene(*args, **kwargs):
            expectation = operator(*args, **kwargs)
            expectation.fill(expectation[element])
            return expectation
        if hasattr(operator, '_bound_onsite'):
            homogene._bound_onsite = operator._bound_onsite
        if hasattr(operator, '_bound_hamiltonian'):
            homogene._bound_hamiltonian = operator._bound_hamiltonian
        return homogene

    # this operator has an array like output, on all sites the value is similar to
    # the density operator
    density_operator_array = make_operator(density_operator_array)

    time = 5
    tmax = time
    params = {'v': 0.0001}
    occupation = manybody.lead_occupation(chemical_potential=3)

    state = manybody.State(syst, tmax, occupation, params=params, refine=False)
    state.evolve(time)

    density = state.evaluate(density_operator, root=None)
    density_array = state.evaluate(density_operator_array, root=None)
    shape_density = density.shape
    shape_density_array = density_array.shape

    # first make sure that the density matches on all points to the scalar value
    for dens in density_array:
        assert np.allclose(dens, density)

    # estimate the error at a given time in the first interval for a given element
    interval = state.get_intervals()[0]

    # ----- a simple quadpack reference implementation
    integrand = ManybodyIntegrand(syst, interval=interval, time=time, params=params,
                                  operator=density_operator)

    tmp = quadpack_qag(integrand.func, interval.kmin, interval.kmax,
                       order=interval.order, atol=1.0e-3, rtol=1.0e-3)

    result_ref, errors_ref, max_error_per_interval_ref, intervals_ref = tmp

    # ----- test for an operator with an array output
    tmp = state.refine_intervals(error_op=density_operator_array, intervals=interval,
                                 atol=1.0e-3, rtol=1.0e-3)
    error_sum, intervals, errors = tmp

    # check the return types
    assert isinstance(error_sum, np.float)
    assert isinstance(intervals, list)
    assert isinstance(errors, np.ndarray)
    assert len(intervals) == len(errors)
    for inter in intervals:
        assert isinstance(inter, manybody.Interval)
    assert errors.shape == (len(intervals), ) + shape_density_array
    max_error = 0
    for error in errors:
        max_error += error
    assert np.allclose(error_sum, np.max(max_error))

    # check that the intervals match
    assert len(intervals) == len(intervals_ref)
    for inte, inte_ref in zip(intervals, intervals_ref):
        assert np.allclose(inte.kmin, inte_ref[0])
        assert np.allclose(inte.kmax, inte_ref[1])

    # check that the errors match
    assert len(errors) == len(errors_ref)
    for err, err_ref in zip(errors, errors_ref):
        assert np.allclose(err[0], err_ref, atol=1e-6)

    # ----- test for an operator with a single scalar output

    # set up a new state, as the old one has refined intervals
    state = manybody.State(syst, tmax, occupation, params=params, refine=False)
    tmp = state.refine_intervals(error_op=density_operator, intervals=interval,
                                 atol=1.0e-3, rtol=1.0e-3)
    error_sum, intervals, errors = tmp

    # check the return types
    assert isinstance(error_sum, np.float)
    assert isinstance(intervals, list)
    assert isinstance(errors, np.ndarray)
    assert len(intervals) == len(errors)
    for inter in intervals:
        assert isinstance(inter, manybody.Interval)
    assert errors.shape == (len(intervals), ) + shape_density
    assert np.allclose(error_sum, np.sum(errors))

    # check that the intervals match
    assert len(intervals) == len(intervals_ref)
    for inte, inte_ref in zip(intervals, intervals_ref):
        assert np.allclose(inte.kmin, inte_ref[0])
        assert np.allclose(inte.kmax, inte_ref[1])

    # check that the errors match
    assert len(errors) == len(errors_ref)
    for err, err_ref in zip(errors, errors_ref):
        assert np.allclose(err, err_ref, atol=1e-6)


@mpi_testdriver.wrap(num_processes)
def test_add_boundstates_to_manybody_state():

    comm = mpi.get_communicator()
    assert comm.size > 1

    syst = make_system().finalized()
    density_operator = kwant.operator.Density(syst)

    tmax = 50
    time = 5
    params = {'v': 0.0001}

    occupation = manybody.lead_occupation(chemical_potential=3)

    # manybody solver
    state = manybody.State(syst, tmax, occupation, params=params, refine=False)

    def make_psi(energy):
        return onebody.ScatteringStates(syst, lead=0, energy=energy,
                                        tmax=tmax, params=params)[0]

    boundstate_tasks = {100: onebody.Task(weight=0.2, energy=2.1, lead=None, mode=None),
                        101: onebody.Task(weight=0.3, energy=2.3, lead=None, mode=None)}

    boundstate_tasks_distributed = mpi.distribute_dict(boundstate_tasks,
                                                       mpi.round_robin,
                                                       comm)

    boundstate_psi = {key: make_psi(task.energy) for key, task
                      in boundstate_tasks_distributed.items()}

    # test simply that we can evolve and evaluate the state
    state.add_boundstates(boundstate_psi, boundstate_tasks)
    state.evolve(time)
    state.evaluate(density_operator)

    # test for error if key are alread occupied
    with pytest.raises(KeyError) as exc:
        state.add_boundstates(boundstate_psi, boundstate_tasks)
    error_string = str(exc.value)
    message = 'boundstate key=100 already present in solver.'
    assert message in error_string

    # test if the input dict is not distributed
    boundstate_tasks = {102: onebody.Task(weight=0.2, energy=2.1, lead=None, mode=None),
                        103: onebody.Task(weight=0.3, energy=2.3, lead=None, mode=None)}
    boundstate_psi = {key: make_psi(task.energy) for key, task
                      in boundstate_tasks.items()}
    raises(AssertionError, state.add_boundstates, boundstate_psi, boundstate_tasks)

    # test if the keys do not match in both input dicts
    boundstate_tasks_distributed = mpi.distribute_dict(boundstate_tasks,
                                                       mpi.round_robin,
                                                       comm)
    boundstate_psi = {key: make_psi(task.energy) for key, task
                      in boundstate_tasks_distributed.items()}
    del boundstate_tasks[103]
    raises(AssertionError, state.add_boundstates, boundstate_psi, boundstate_tasks)


@mpi_testdriver.wrap(num_processes)
def test_get_rank():

    rank = mpi.get_rank()
    comm = mpi.get_communicator()

    assert isinstance(rank, dict)
    assert rank['rank'] == comm.rank
